# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Implements adapter for lifelines models."""

__all__ = ["_LifelinesAdapter"]
__author__ = ["fkiraly"]

from warnings import warn

import numpy as np
import pandas as pd

from skpro.distributions.empirical import Empirical
from skpro.survival.adapters._common import _clip_surv, _get_fitted_params_default_safe
from skpro.utils.sklearn import prep_skl_df


class _LifelinesAdapter:
    """Mixin adapter class for lifelines models."""

    _tags = {
        # packaging info
        # --------------
        "authors": ["fkiraly"],
        "python_dependencies": ["lifelines"],
        "license_type": "permissive",
        # capability tags
        # ---------------
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
        "C_inner_mtype": "pd_DataFrame_Table",
        "capability:multioutput": False,
    }

    # defines the name of the attribute containing the lifelines estimator
    _estimator_attr = "_estimator"

    def _get_lifelines_class(self):
        """Abstract method to get lifelines class.

        should import and return lifelines class
        """
        # from lifelines import LifelinesClass
        #
        # return LifelinesClass
        raise NotImplementedError("abstract method")

    def _get_lifelines_object(self):
        """Abstract method to initialize lifelines object.

        The default initializes result of _get_lifelines_class
        with self.get_params.
        """
        cls = self._get_lifelines_class()
        return cls(**self.get_params())

    def _init_lifelines_object(self):
        """Abstract method to initialize lifelines object and set to _estimator_attr.

        The default writes the return of _get_lifelines_object to
        the attribute of self with name _estimator_attr
        """
        cls = self._get_lifelines_object()
        setattr(self, self._estimator_attr, cls)
        return getattr(self, self._estimator_attr)

    def _get_extra_fit_args(self, X, y, C=None):
        """Get extra arguments for the fit method.

        Parameters
        ----------
        X : pd.DataFrame
            Training features
        y: pd.DataFrame
            Training labels
        C: pd.DataFrame, optional (default=None)
            Censoring information for survival analysis.

        Returns
        -------
        dict
            Extra arguments for the fit method.
        """
        return {}

    def _fit(self, X, y, C=None):
        """Fit estimator training data.

        Parameters
        ----------
        X : pd.DataFrame
            Training features
        y: pd.DataFrame
            Training labels
        C: pd.DataFrame, optional (default=None)
            Censoring information for survival analysis.

        Returns
        -------
        self: reference to self
            Fitted estimator.
        """
        lifelines_est = self._init_lifelines_object()

        # input conversion
        X = X.astype("float")  # lifelines insists on float dtype
        X = prep_skl_df(X)

        if hasattr(self, "X_col_subset"):
            X = X[self.X_col_subset]

        to_concat = [X, y]

        if C is not None:
            C_col = 1 - C.copy()  # lifelines uses 1 for uncensored, 0 for censored
            C_col.columns = ["__C"]
            to_concat.append(C_col)

        df = pd.concat(to_concat, axis=1)

        self._y_cols = y.columns  # remember column names for later
        y_name = y.columns[0]

        fit_args = {
            "df": df,
            "duration_col": y_name,
        }
        if C is not None:
            fit_args["event_col"] = "__C"

        fit_args.update(self._get_extra_fit_args(X, y, C))

        # fit lifelines estimator
        lifelines_est.fit(**fit_args)

        # write fitted params to self
        # some fitted parameters are properties and may raise exceptions
        # for example, AIC_ and AIC_partial_ of CoxPHFitter
        # to avoid this, we use a safe getter
        lifelines_fitted_params = _get_fitted_params_default_safe(lifelines_est)
        for k, v in lifelines_fitted_params.items():
            setattr(self, f"{k}_", v)

        return self

    def _predict_proba(self, X):
        """Predict_proba method adapter.

        Parameters
        ----------
        X : pd.DataFrame
            Features to predict on.

        Returns
        -------
        skpro Empirical distribution
        """
        lifelines_est = getattr(self, self._estimator_attr)

        # input conversion
        X = X.astype("float")  # lifelines insists on float dtype
        X = prep_skl_df(X)

        # predict on X
        lifelines_survf = lifelines_est.predict_survival_function(X)

        times = lifelines_survf.index

        nt = len(times)
        mi = pd.MultiIndex.from_product([X.index, range(nt)]).swaplevel()

        times_val = np.repeat(times, repeats=len(X))
        times_df = pd.DataFrame(times_val, index=mi, columns=self._y_cols)

        lifelines_survf_t = np.transpose(lifelines_survf.values)
        _, lifelines_survf_t_diff, clipped = _clip_surv(lifelines_survf_t)

        if clipped:
            warn(
                f"Warning from {self.__class__.__name__}: "
                f"Interfaced lifelines class {lifelines_est.__class__.__name__} "
                "produced improper survival function predictions, i.e., "
                "not monotonically decreasing or not in [0, 1]. "
                "skpro has clipped the predictions to enforce proper range and "
                "valid predictive distributions. "
                "However, predictions may still be unreliable.",
                stacklevel=2,
            )

        weights = -lifelines_survf_t_diff.flatten()
        weights_df = pd.Series(weights, index=mi)

        dist = Empirical(
            spl=times_df, weights=weights_df, index=X.index, columns=self._y_cols
        )

        return dist



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

"""Skew-Normal probability distribution."""

__author__ = ["Spinachboul"]

from scipy.stats import skewnorm

from skpro.distributions.adapters.scipy._distribution import _ScipyAdapter


class SkewNormal(_ScipyAdapter):
    r"""Skew-Normal Probability Distribution.

    The skew-normal distribution generalizes the normal distribution by introducing
    a shape parameter :math:`\\alpha` to control skewness. It is parameterized by
    ``mu``, ``sigma``, and ``alpha``:

    .. math:: f(x; \alpha, \\mu, \\sigma) = \frac{2}{\\sigma}
          \\phi\\left(\frac{x - \\mu}{\\sigma}\right)
          \\Phi\\left(\alpha \frac{x - \\mu}{\\sigma}\right)

    where :math:`\\phi` and :math:`\\Phi` are the pdf and cdf of the standard normal
    distribution, respectively, :math:`\\mu` is the location parameter,
    :math:`\\sigma` is the scale parameter (must be positive), and
    :math:`\alpha` is the shape parameter controlling skewness.

    Parameters
    ----------
    mu : float
        Location parameter of the distribution (mean shift).
    sigma : float
        Scale parameter of the distribution (standard deviation), must be positive.
    alpha : float
        Skewness parameter of the distribution:
        - A positive value skews to the right.
        - A negative value skews to the left.
        - A value of zero results in a standard normal distribution.
    index : array-like, optional, default=None
        Index for the distribution, providing pandas-like indexing for rows.
    columns : array-like, optional, default=None
        Columns for the distribution, providing pandas-like indexing for columns.

    Examples
    --------
    >>> from skpro.distributions import SkewNormal
    >>> s = SkewNormal(mu=0, sigma=1, alpha=3)
    """

    _tags = {
        "authors": ["Spinachboul"],
        "python_dependencies": ["scipy"],
        "distr:measuretype": "continuous",
        "capabilities:approx": ["energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "broadcast_init": "on",
    }

    def __init__(self, mu, sigma, alpha=0, index=None, columns=None):
        self.mu = mu
        self.sigma = sigma
        self.alpha = alpha
        self.index = index
        self.columns = columns

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self):
        return skewnorm

    def _get_scipy_param(self):
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]
        alpha = self._bc_params["alpha"]

        return [], {
            "loc": mu,
            "scale": sigma,
            "a": alpha,
        }

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        r"""Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict
            Parameters to create testing instances of the class. Each dict represents
            parameters to construct a test instance:
            - `MyClass(**params)` or `MyClass(**params[i])` creates a valid test.
            - `create_test_instance` uses the first (or only) dictionary in `params`.
        """
        return [
            {"mu": -1, "sigma": 2},
            {"mu": 0, "sigma": 1, "alpha": -2},
        ]



"""Interface for xgboostlss probabilistic regressor."""

from skpro.regression.base import BaseProbaRegressor


class XGBoostLSS(BaseProbaRegressor):
    """Interface to xgboostlss regerssor from the xgboostlss package.

    Direct interface to ``XGBoostLSS`` from ``xgboostlss`` package by ``StatMixedML``.

    Parameters
    ----------
    dist: str, optional, default="Normal"
        Form of predictive distribution, strings are same as in skpro.

        Valid options are:

        * "Normal": Normal distribution.
        * "Gamma": Gamma distribution.
        * "Laplace": Laplace distribution.
        * "LogNormal": LogNormal distribution.
        * "TDistribution": Student's T distribution.
        * "Weibull": Weibull distribution.

    stabilization: str, optional, default="None"
        Stabilization method for the Gradient and Hessian.
        Options are "None", "MAD", "L2".

    response_fn: str, optional, default="exp"
        Response function for transforming the distributional parameters to the
        support of the distribution. Options are "exp" (exponential) or
        "softplus" (softplus).

    loss_fn: str, optional, default="nll"
        Loss function used in tuning and fitting.
        Options are "nll" (negative log-likelihood)
        or "crps" (continuous ranked probability score).

    n_cpu: int or str, optional, default="auto"
        Number of CPUs to use for parallel processing of data in ``xgboostlss``.
        Default is "auto" which uses all available CPUs.

    num_boost_round: int, optional, default=100
        Number of boosting iterations.

    nfold: int, optional, default=5
        Number of folds in CV used for tuning.

    early_stopping_rounds: int, optional, default=20
        Number of early stopping round interval.
        Cross-Validation metric (average of validation metric computed over CV folds)
        needs to improve at least once every **early_stopping_rounds**
        round(s) to continue training.
        The last entry in the evaluation history will represent the best iteration.

    max_minutes: int, optional, default=10
        Time budget in minutes, i.e., stop study after the given number of minutes.

    n_trials: int, optional, default=30
        The number of trials in tuning.
        If this argument is set to None, there is no limitation on the number of trials.
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["StatMixedML", "EchoDel", "fkiraly"],
        # StatMixedML, EchoDel for the original xgboostlss package
        "maintainers": ["fkiraly"],
        "python_dependencies": ["xgboostlss"],  # PEP 440 python dependencies specifier,
        #
        # estimator tags
        # --------------
        "capability:multioutput": False,  # can the estimator handle multi-output data?
        "capability:missing": True,  # can the estimator handle missing data?
        "X_inner_mtype": "pd_DataFrame_Table",  # type seen in internal _fit, _predict
        "y_inner_mtype": "pd_DataFrame_Table",  # type seen in internal _fit
    }

    def __init__(
        self,
        dist="Normal",
        stabilization="None",
        response_fn="exp",
        loss_fn="nll",
        n_cpu="auto",
        num_boost_round=100,
        nfold=5,
        early_stopping_rounds=20,
        max_minutes=10,
        n_trials=30,
    ):
        self.dist = dist
        self.stabilization = stabilization
        self.response_fn = response_fn
        self.loss_fn = loss_fn
        self.n_cpu = n_cpu
        self.num_boost_round = num_boost_round
        self.nfold = nfold
        self.early_stopping_rounds = early_stopping_rounds
        self.max_minutes = max_minutes
        self.n_trials = n_trials

        super().__init__()

        if n_cpu == "auto":
            import multiprocessing

            self._n_cpu = multiprocessing.cpu_count()
        else:
            self._n_cpu = n_cpu

    def _get_xgblss_distr(self, distr):
        """Get xgboostlss distribution object from string.

        Parameters
        ----------
        distr : str
            Distribution name, in skpro, as in self.dist
        """
        import importlib

        SKPRO_TO_XGBLSS = {
            "Normal": "Gaussian",
            "TDistribution": "StudentT",
        }
        distr = SKPRO_TO_XGBLSS.get(distr, distr)

        module_str = "xgboostlss.distributions." + distr
        object_str = distr

        module = importlib.import_module(module_str)
        return getattr(module, object_str)

    def _get_skpro_distr(self, distr):
        """Get skpro distribution object from string.

        Parameters
        ----------
        distr : str
            Distribution name, in skpro, as in self.dist
        """
        import importlib

        module_str = "skpro.distributions"
        object_str = distr

        module = importlib.import_module(module_str)
        return getattr(module, object_str)

    def _get_skpro_val_dict(self, distr, df):
        """Convert xgboostlss parameters to skpro distribution.

        Parameters
        ----------
        distr : str
            Distribution name, in skpro, as in self.dist
        df : pd.DataFrame
            DataFrame of parameters as returned by predict, in xgboostlss.
        """
        name_map = {
            "Normal": {"mu": "loc", "sigma": "scale"},
            "Gamma": {"alpha": "concentration", "beta": "rate"},
            "Laplace": {"mu": "loc", "scale": "scale"},
            "LogNormal": {"mu": "loc", "sigma": "scale"},
            "TDistribution": {"mu": "loc", "sigma": "scale", "df": "df"},
            "Weibull": {"scale": "scale", "k": "concentration"},
        }
        map = name_map.get(distr, {})

        vals = {k: df.loc[:, [v]].values for k, v in map.items()}

        return vals

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        import xgboost as xgb
        from xgboostlss.model import XGBoostLSS

        self._y_cols = y.columns
        n_cpu = self._n_cpu

        dtrain = xgb.DMatrix(X, label=y, nthread=n_cpu, silent=True)

        xgblss_distr = self._get_xgblss_distr(self.dist)

        xgblss = XGBoostLSS(
            xgblss_distr(
                stabilization="None",
                response_fn="exp",
                loss_fn="nll",
            )
        )

        param_dict = {
            "eta": ["float", {"low": 1e-5, "high": 1, "log": True}],
            "max_depth": ["int", {"low": 1, "high": 10, "log": False}],
            "gamma": ["float", {"low": 1e-8, "high": 40, "log": True}],
            "subsample": ["float", {"low": 0.2, "high": 1.0, "log": False}],
            "colsample_bytree": ["float", {"low": 0.2, "high": 1.0, "log": False}],
            "min_child_weight": ["float", {"low": 1e-8, "high": 500, "log": True}],
            "booster": ["categorical", ["gbtree"]],
        }

        opt_param = xgblss.hyper_opt(
            param_dict,
            dtrain,
            num_boost_round=self.num_boost_round,
            # Number of boosting iterations.
            nfold=self.nfold,
            # Number of cv-folds.
            early_stopping_rounds=self.early_stopping_rounds,
            # Number of early-stopping rounds
            max_minutes=self.max_minutes,
            # Time budget in minutes,
            # i.e., stop study after the given number of minutes.
            n_trials=self.n_trials,
            # The number of trials. If this argument is set to None,
            # there is no limitation on the number of trials.
            silence=True,
            # Controls the verbosity of the trail,
            # i.e., user can silence the outputs of the trail.
            seed=123,
            # Seed used to generate cv-folds.
            hp_seed=123,
            # Seed for random number generator used
            # in the Bayesian hyperparameter search.
        )

        opt_params = opt_param.copy()
        n_rounds = opt_params["opt_rounds"]
        del opt_params["opt_rounds"]

        # Train Model with optimized hyperparameters
        xgblss.train(opt_params, dtrain, num_boost_round=n_rounds)

        self.xgblss_ = xgblss
        return self

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        import xgboost as xgb

        n_cpu = self._n_cpu

        index = X.index
        y_cols = self._y_cols
        columns = y_cols

        dtest = xgb.DMatrix(X, nthread=n_cpu, silent=True)

        y_pred_xgblss = self.xgblss_.predict(dtest, pred_type="parameters")

        skpro_distr = self._get_skpro_distr(self.dist)
        skpro_distr_vals = self._get_skpro_val_dict(self.dist, y_pred_xgblss)

        y_pred = skpro_distr(**skpro_distr_vals, index=index, columns=columns)

        return y_pred

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params0 = {"max_minutes": 1}
        params1 = {
            "stabilization": "L2",
            "loss_fn": "crps",
            "max_minutes": 1,
        }
        params2 = {"dist": "Gamma", "max_minutes": 1}
        params3 = {"dist": "Weibull", "max_minutes": 1}
        params4 = {"dist": "TDistribution", "max_minutes": 1}
        params5 = {"dist": "Laplace", "max_minutes": 1}
        return [params0, params1, params2, params3, params4, params5]



"""Module exports: Series type checkers, converters and mtype inference."""

from skpro.datatypes._table._convert import convert_dict as convert_dict_Table
from skpro.datatypes._table._registry import MTYPE_LIST_TABLE, MTYPE_REGISTER_TABLE

__all__ = [
    "convert_dict_Table",
    "MTYPE_LIST_TABLE",
    "MTYPE_REGISTER_TABLE",
]



"""Probability distribution objects."""

# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__all__ = [
    "Alpha",
    "Beta",
    "Binomial",
    "ChiSquared",
    "Delta",
    "Empirical",
    "Erlang",
    "Exponential",
    "Fisk",
    "Gamma",
    "HalfCauchy",
    "HalfLogistic",
    "HalfNormal",
    "IID",
    "InverseGamma",
    "Histogram",
    "Laplace",
    "Logistic",
    "LogLaplace",
    "LogNormal",
    "MeanScale",
    "Mixture",
    "Normal",
    "Pareto",
    "Poisson",
    "QPD_Empirical",
    "QPD_S",
    "QPD_B",
    "QPD_U",
    "QPD_Johnson",
    "SkewNormal",
    "TDistribution",
    "TruncatedNormal",
    "Uniform",
    "Weibull",
]

from skpro.distributions.alpha import Alpha
from skpro.distributions.beta import Beta
from skpro.distributions.binomial import Binomial
from skpro.distributions.chi_squared import ChiSquared
from skpro.distributions.compose import IID
from skpro.distributions.delta import Delta
from skpro.distributions.empirical import Empirical
from skpro.distributions.erlang import Erlang
from skpro.distributions.exponential import Exponential
from skpro.distributions.fisk import Fisk
from skpro.distributions.gamma import Gamma
from skpro.distributions.halfcauchy import HalfCauchy
from skpro.distributions.halflogistic import HalfLogistic
from skpro.distributions.halfnormal import HalfNormal
from skpro.distributions.histogram import Histogram
from skpro.distributions.inversegamma import InverseGamma
from skpro.distributions.laplace import Laplace
from skpro.distributions.logistic import Logistic
from skpro.distributions.loglaplace import LogLaplace
from skpro.distributions.lognormal import LogNormal
from skpro.distributions.meanscale import MeanScale
from skpro.distributions.mixture import Mixture
from skpro.distributions.normal import Normal
from skpro.distributions.pareto import Pareto
from skpro.distributions.poisson import Poisson
from skpro.distributions.qpd import QPD_B, QPD_S, QPD_U, QPD_Johnson
from skpro.distributions.qpd_empirical import QPD_Empirical
from skpro.distributions.skew_normal import SkewNormal
from skpro.distributions.t import TDistribution
from skpro.distributions.truncated_normal import TruncatedNormal
from skpro.distributions.uniform import Uniform
from skpro.distributions.weibull import Weibull



"""Common utilities for adapters."""

import numpy as np


def _clip_surv(surv_arr):
    """Clips improper survival function values to proper range.

    Enforces: values are in [0, 1] and are monotonically decreasing.

    First clips to [0, 1], then enforces monotonicity, by replacing
    any value with minimum of itself and any previous values.

    Parameters
    ----------
    surv_arr : 2D np.ndarray
        Survival function values.
        index 0 is instance index.
        index 1 is time index, increasing.

    Returns
    -------
    surv_arr_clipped : 2D np.ndarray
        Clipped survival function values.
    surv_arr_diff : 2D np.ndarray, same shape as surv_arr_clipped.
        Difference of clipped survival function values.
        Same as np.diff(surv_arr, axis=1, prepend=1, append=0),
        then summing the last two columns to become one column.
        Returned to avoid recomputation, if needed later in context.
    clipped : boolean
        Whether clipping was needed.
    """
    too_large = surv_arr > 1
    too_small = surv_arr < 0

    surv_arr[too_large] = 1
    surv_arr[too_small] = 0

    surv_arr_diff = _surv_diff(surv_arr)

    # avoid iterative minimization if no further clipping is needed
    if not (surv_arr_diff > 0).any():
        clipped = too_large.any() or too_small.any()
        return surv_arr, surv_arr_diff, clipped

    # enforce monotonicity
    # iterating from left to right ensures values are replaced
    # with minimum of itself and all values to the left
    for i in range(1, surv_arr.shape[1]):
        surv_arr[:, i] = np.minimum(surv_arr[:, i], surv_arr[:, i - 1])

    surv_arr_diff = _surv_diff(surv_arr)

    return surv_arr, surv_arr_diff, True


def _surv_diff(surv_arr):
    """Compute difference of survival function values.

    Parameters
    ----------
    surv_arr : 2D np.ndarray
        Survival function values.
        index 0 is instance index.
        index 1 is time index, increasing.

    Returns
    -------
    surv_arr_diff : 2D np.ndarray, same shape as surv_arr
        Difference of survival function values.
        Same as np.diff(surv_arr, axis=1, prepend=1, append=0),
        then summing the last two columns to become one column
    """
    surv_arr_diff = np.diff(surv_arr, axis=1, prepend=1, append=0)

    surv_arr_diff[:, -2] = surv_arr_diff[:, -2] + surv_arr_diff[:, -1]
    surv_arr_diff = surv_arr_diff[:, :-1]

    return surv_arr_diff


def _get_fitted_params_default_safe(obj=None):
    """Obtain fitted params of object, per sklearn convention.

    Same as _get_fitted_params_default, but with exception handling.

    This is since in sksurv, feature_importances_ is a property
    and may raise an exception if the estimator does not have it.

    Parameters
    ----------
    obj : any object

    Returns
    -------
    fitted_params : dict with str keys
        fitted parameters, keyed by names of fitted parameter
    """
    # default retrieves all self attributes ending in "_"
    # and returns them with keys that have the "_" removed
    #
    # get all attributes ending in "_", exclude any that start with "_" (private)
    fitted_params = [
        attr for attr in dir(obj) if attr.endswith("_") and not attr.startswith("_")
    ]

    def hasattr_safe(obj, attr):
        try:
            if hasattr(obj, attr):
                getattr(obj, attr)
                return True
        except Exception:
            return False

    # remove the "_" at the end
    fitted_param_dict = {
        p[:-1]: getattr(obj, p) for p in fitted_params if hasattr_safe(obj, p)
    }

    return fitted_param_dict



"""Probability distribution objects."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__all__ = ["IID"]

from skpro.distributions.compose._iid import IID



"""Concordance index, Harrell's."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

import numpy as np
import pandas as pd

from skpro.metrics.base import BaseSurvDistrMetric


class ConcordanceHarrell(BaseSurvDistrMetric):
    r"""Concordance index (Harrell).

    Fraction of concordant test index pairs among all comparale pairs,
    as proposed in [1]_, commonly known as Harrell's C-index, Harrell's C,
    or simply concordance index,
    if not in delination of other C-indices (e.g., Uno's C-index).

    For ground truth samples :math:`y_i, c_i, i = 1 \dots N`,
    and predicted inverse risk scores :math:`s_i, i = 1 \dots N`,
    a pair of test non-equal test indices :math:`i \lneq j` is concordant if
    :math:`(y_i > y_j) \land (s_i > s_j)` or :math:`(y_i < y_j) \land (s_i < s_j)`.
    If :math:`(s_i = s_j)`, the pair is counted as concordant if :math:`y_i = y_j`,
    and :math:`c_i = c_j = 0`, otherwise it is considered a tie,
    counted as half concordant, half discordant by default.

    A pair of test indices :math:`i \lneq j` is said to be comparable
    if one of the following conditions holds:

    * :math:`y_i > y_j` and :math:`c_j = 0`
    * :math:`y_i < y_j` and :math:`c_i = 0`
    * :math:`y_i = y_j` and :math:`c_i c_j = 0`

    This metric supports multiple options for inverse risk scores,
    including any method evaluates of predictive distributions.

    The default is the predictive mean survival time.

    `evaluate` computes the concordance index.
    `evaluate_by_index` produces, for one test sample,
    the fraction of concordant pairs among all pairs with this sample as first index.
    `multivariate` controls averaging over variables.

    Parameters
    ----------
    score : str, optional, default='mean'
        The type of inverse risk score to use.
        Calls predict_proba, then the method of the same name as `score`.
        Examples include 'mean', 'median', 'quantile', 'cdf'.

    score_args : dict, optional, default=None
        Additional arguments to pass to the score method, e.g., quantiles.

    higher_score_is_lower_risk : bool, optional, default=True
        If True, higher score is considered lower risk, and vice versa,
        that is, the score is assumed to be an inverse risk score.
        If False, the score is assumed to be a risk score, and a
        negative sign is applied to the score.

    tie_score : float, optional, default=0.5
        The value to use for ties in the risk scores,
        as a relative value to counting as concordant.
        1 is counting as concordant, 0 is counting as discordant.
        0.5 is counting as half concordant, half discordant.

    normalization : str, {'overall', 'index'}, optional, default='overall'
        Determines the normalization of the concordance index, whether
        fractions of concordant pairs are averaged primarily overall,
        or primarily per index. In both cases, ``evaluate`` returns the
        arithmetic mean of ``evaluate_by_index``.

        * If ``'overall'``, ``evaluate``
        returns the fraction of concordant among all comparable pairs.
        This is as in [1]_.
        In ``evaluate_by_index``, fraction denominators are the
        number of comparable pairs overall, divided by the number of samples,
        instead of the number of comparable pairs in which the index is the first index.

        * If ``'index'``, ``evaluate`` returns the average, over indices,
        of the fraction of concordant pairs among
        all comparable pairs in which the index is the first index.
        In ``evaluate_by_index``, entries are the fraction of concordant pairs
        among all comparable pairs in which the index is the first index,
        without further normalization.

    multioutput : {'raw_values', 'uniform_average'} or array-like of shape
        (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.
        If 'uniform_average' (default), errors are mean-averaged across variables.
        If array-like, errors are weighted averaged across variables, values as weights.
        If 'raw_values', does not average errors across variables, columns are retained.
    multivariate : bool, optional, default=False
        if True, behaves as multivariate log-loss
        C-index is computed for entire row, results one score per row
        if False, is univariate log-loss
        C-index is computed per variable marginal, results in many scores per row

    References
    ----------
    .. [1] Harrell FE, Califf R.M., Pryor DB, Lee KL, Rosati RA.
      Multivariable prognostic models: issues in developing models,
      evaluating assumptions and adequacy, and measuring and reducing errors.
      Statistics in Medicine, 15(4), 361-87, 1996.
    """

    _tags = {
        "authors": "fkiraly",
        "capability:survival": True,
        "scitype:y_pred": "pred_proba",
        "lower_is_better": False,
    }

    def __init__(
        self,
        score="mean",
        score_args=None,
        higher_score_is_lower_risk=True,
        tie_score=0.5,
        normalization="overall",
        multioutput="uniform_average",
        multivariate=False,
    ):
        self.score = score
        self.score_args = score_args
        self.higher_score_is_lower_risk = higher_score_is_lower_risk
        self.tie_score = tie_score
        self.normalization = normalization
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        C_true = kwargs.get("C_true", None)

        # retrieve parameters
        tie_score = self.tie_score
        normalization = self.normalization
        score_args = self.score_args
        if score_args is None:
            score_args = {}

        # convert to numpy and remember index and columns
        ix = y_true.index
        cols = y_true.columns

        y_true = y_true.to_numpy()
        if C_true is not None:
            C_true = C_true.to_numpy()
        else:
            C_true = np.zeros_like(y_true)

        risk_scores = getattr(y_pred, self.score)(**score_args)
        if not self.higher_score_is_lower_risk:
            risk_scores = -risk_scores
        risk_scores = risk_scores.to_numpy()

        ncomp_mat = np.zeros_like(y_true)
        nconc_mat = np.zeros_like(y_true)
        result = np.zeros_like(y_true)

        # compute concordance index for each index
        for j in range(y_true.shape[1]):
            yj = y_true[:, j]
            Cj = C_true[:, j] == 1
            nCj = ~Cj
            rj = risk_scores[:, j]
            for i in range(y_true.shape[0]):
                yij = yj[i]
                rij = rj[i]
                Cij = Cj[i]
                nCij = ~Cij
                one_unc = ~(Cj & Cij)  # at least one uncensored in the pair
                xone_unc = ~(Cj & Cij) & (Cj | Cij)  # exactly one uncensored

                # mark concordant pairs (no ties)
                comp1 = nCij & (yj > yij)  # comparable, > type
                conc1 = comp1 & (rj > rij)
                comp2 = nCj & (yj < yij)  # comparable, < type
                conc2 = comp2 & (rj < rij)

                # mark concordant pairs, handling ties
                conc3 = nCij & nCj & ((yj == yij) & (rj == rij))
                conc = conc1 | conc2 | conc3

                # count concordant pairs
                nconc = conc.sum() - nCij  # sum, subtract i=j if counted above
                # i=j was counted iff it was not censored

                # handle ties in total of concordant pairs
                if tie_score != 0:
                    nconc = nconc.astype(float)
                    nconc += np.sum((yj != yij) & (rj == rij)) * tie_score
                    nconc += np.sum(xone_unc & (yj == yij) & (rj == rij)) * tie_score

                # count comparable pairs
                comp3 = one_unc & (yj == yij)
                comp = comp1 | comp2 | comp3
                ncomp = comp.sum() - nCij  # subtract i=j, but only if counted above

                nconc_mat[i, j] = nconc
                ncomp_mat[i, j] = ncomp

        # normalization
        if normalization == "overall":
            # weighting is such that average over rows
            # results in the overall C-index
            ncomp_total = ncomp_mat.sum(axis=0)
            nspl = len(ncomp_mat)
            result = (nspl / ncomp_total) * nconc_mat
        else:  # normalization == "index"
            # weighting is such that rows contain simple fractions
            # but the average over rows is not the overall C-index,
            # as number of comparable pairs is in general not the same for each index
            ncomp_mat[ncomp_mat == 0] = 1  # handling of case 0 / 0, avoid nans
            result = nconc_mat / ncomp_mat
            result[ncomp_mat == 1] = tie_score  # handling of case 0 / 0

        res_df = pd.DataFrame(result, index=ix, columns=cols)

        if self.multivariate:
            return pd.DataFrame(res_df.mean(axis=1), columns=["C_Harrell"])
        else:
            return res_df

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {}
        params2 = {"score": "quantile", "score_args": {"alpha": 0.5}}
        params3 = {"normalization": "index"}

        return [params1, params2, params3]



"""Tests for probabilistic metrics for distribution predictions."""
import numpy as np
import pandas as pd
import pytest
from skbase.testing import QuickTester

from skpro.distributions import Normal
from skpro.tests.test_all_estimators import BaseFixtureGenerator, PackageConfig

TEST_DISTS = [Normal]


class TestAllDistrMetrics(PackageConfig, BaseFixtureGenerator, QuickTester):
    """Generic tests for all probabilistic regression metrics in the package."""

    # class variables which can be overridden by descendants
    # ------------------------------------------------------

    # which object types are generated; None=all, or scitype string
    # passed to skpro.registry.all_objects as object_type
    object_type_filter = "metric_distr"

    @pytest.mark.parametrize("dist", TEST_DISTS)
    @pytest.mark.parametrize("pass_c", [True, False])
    @pytest.mark.parametrize("multivariate", [True, False])
    @pytest.mark.parametrize("multioutput", ["raw_values", "uniform_average"])
    def test_distr_evaluate(
        self, object_instance, dist, pass_c, multivariate, multioutput
    ):
        """Test expected output of evaluate functions."""
        metric = object_instance

        y_pred = dist.create_test_instance()
        y_true = y_pred.sample()

        m = metric.set_params(multioutput=multioutput)
        if "multivariate" in metric.get_params():
            m = m.set_params(multivariate=multivariate)

        if not multivariate:
            expected_cols = y_true.columns
        else:
            expected_cols = ["score"]

        metric_args = {"y_true": y_true, "y_pred": y_pred}
        if pass_c:
            c_true = np.random.randint(0, 2, size=y_true.shape)
            c_true = pd.DataFrame(c_true, columns=y_true.columns, index=y_true.index)
            metric_args["c_true"] = c_true

        res = m.evaluate_by_index(**metric_args)
        assert isinstance(res, pd.DataFrame)
        assert (res.columns == expected_cols).all()
        assert res.shape == (y_true.shape[0], len(expected_cols))

        res = m.evaluate(**metric_args)

        expect_df = not multivariate and multioutput == "raw_values"
        if expect_df:
            assert isinstance(res, pd.DataFrame)
            assert (res.columns == expected_cols).all()
            assert res.shape == (1, len(expected_cols))
        else:
            assert isinstance(res, float)



"""Base class and template for regressors and transformers."""
from skbase.base import BaseEstimator as _BaseEstimator
from skbase.base import BaseMetaEstimator as _BaseMetaEstimator
from skbase.base import BaseObject as _BaseObject


class _CommonTags:
    """Mixin for common tag definitions to all estimator base classes."""

    # config common to all estimators
    _config = {}

    _tags = {
        "estimator_type": "estimator",
        "authors": "skpro developers",
        "maintainers": "skpro developers",
    }

    @property
    def name(self):
        """Return the name of the object or estimator."""
        return self.__class__.__name__


class BaseObject(_CommonTags, _BaseObject):
    """Base class for fittable objects."""

    def __init__(self):
        super().__init__()


class BaseEstimator(_CommonTags, _BaseEstimator):
    """Base class for fittable objects."""


class BaseMetaEstimator(_CommonTags, _BaseMetaEstimator):
    """Base class for fittable composite meta-objects."""



"""Adapters for probability distribution objects, scipy facing."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.distributions.adapters.scipy._distribution import _ScipyAdapter
from skpro.distributions.adapters.scipy._empirical import empirical_from_discrete

__all__ = ["empirical_from_discrete", "_ScipyAdapter"]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Delta (constant/certain) probability distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Delta(BaseDistribution):
    r"""Delta distribution aka constant distribution aka certain distribution.

    This distribution always produces the same value when sampling - ``c``.
    It it useful to represent a constant value as a distribution, e.g., as a baseline
    method to create a probabilistic prediction from a point prediction.

    The delta distribution is parametrized by a constant value :math:`c`.
    For the cdf, we have:

    .. math:: F(x) = 0 \text{ if } x < c, 1 \text{ if } x \geq c

    The constant value :math:`c` is represented by the parameter ``c``.

    Parameters
    ----------
    c : float or array of float (1D or 2D)
        support of the delta distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.delta import Delta

    >>> delta = Delta(c=[[0, 1], [2, 3], [4, 5]])
    >>> this_is_always_c = delta.sample()
    """

    _tags = {
        "capabilities:approx": [],
        "capabilities:exact": ["mean", "var", "energy", "pmf", "log_pmf", "cdf", "ppf"],
        "distr:measuretype": "discrete",
        "distr:paramtype": "nonparametric",
        "broadcast_init": "on",
    }

    def __init__(self, c, index=None, columns=None):
        self.c = c
        if index is None and hasattr(c, "index") and isinstance(c.index, pd.Index):
            index = c.index
        if columns is None and hasattr(c, "columns"):
            if isinstance(c.columns, pd.Index):
                columns = c.columns

        super().__init__(index=index, columns=columns)

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        # energy of self w.r.t. self is always 0
        energy_arr = self._coerce_to_self_index_np(0)
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        c = self._bc_params["c"]

        energy_arr = np.abs(c - x)
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._bc_params["c"]

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        # variance of a constant is always 0
        return self._coerce_to_self_index_np(0)

    def _pmf(self, x):
        """Probability mass function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pmf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pmf values at the given points
        """
        c = self._bc_params["c"]
        pmf_arr = np.where(x == c, 1, 0)
        return pmf_arr

    def _log_pmf(self, x):
        """Logarithmic probability mass function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pmf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pmf values at the given points
        """
        c = self._bc_params["c"]
        lpmf_arr = np.where(x == c, 0, -np.inf)
        return lpmf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        c = self._bc_params["c"]

        cdf_arr = np.where(x < c, 0, 1)
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        c = self._bc_params["c"]
        icdf_arr = c
        return icdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"c": [[0, 1], [2, 3], [4, 5]]}
        params2 = {
            "c": 42,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"c": 42}
        return [params1, params2, params3]



"""Utility functions for working with indices."""

import numpy as np


def random_ss_ix(ix, size, replace=True):
    """Randomly uniformly sample indices from a list of indices.

    Parameters
    ----------
    ix : pd.Index or subsettable iterable via getitem
        list of indices to sample from
    size : int
        number of indices to sample
    replace : bool, default=True
        whether to sample with replacement
    """
    a = range(len(ix))
    ixs = ix[np.random.choice(a, size=size, replace=replace)]
    return ixs



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Mixture distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd
from skbase.base import BaseMetaObject

from skpro.distributions.base import BaseDistribution


class Mixture(BaseMetaObject, BaseDistribution):
    """Mixture of distributions.

    Parameters
    ----------
    distributions : list of tuples (str, BaseDistribution) or BaseDistribution
        list of mixture components
    weights : list of float, optional, default = None
        list of mixture weights, will be normalized to sum to 1
        if not provided, uniform mixture is assumed
    indep_rows : bool, optional, default = True
        if True, rows are sampled independently from the mixture components.
        If False, the same component is used for all rows.
        Relevant only in ``sample`` method and non-marginal outputs.
    indep_cols : bool, optional, default = True
        if True, columns are sampled independently from the mixture components.
        If False, the same component is used for all columns.
        Relevant only in ``sample`` method and non-marginal outputs.
    index : pd.Index, optional, default = inferred from component distributions
    columns : pd.Index, optional, default = inferred from component distributions

    Examples
    --------
    >>> from skpro.distributions.mixture import Mixture
    >>> from skpro.distributions.normal import Normal

    >>> n1 = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)
    >>> n2 = Normal(mu=3, sigma=2, index=n1.index, columns=n1.columns)
    >>> m = Mixture(distributions=[("n1", n1), ("n2", n2)], weights=[0.3, 0.7])
    >>> mixture_sample = m.sample(n_samples=10)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm", "energy", "ppf"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf"],
        "distr:measuretype": "mixed",
        "distr:paramtype": "composite",
        "named_object_parameters": "_distributions",
    }

    def __init__(
        self,
        distributions,
        weights=None,
        indep_rows=True,
        indep_cols=True,
        index=None,
        columns=None,
    ):
        self.distributions = distributions
        self.weights = weights
        self.indep_rows = indep_rows
        self.indep_cols = indep_cols

        self._distributions = self._coerce_to_named_object_tuples(distributions)
        n_dists = len(self._distributions)

        if weights is None:
            self._weights = np.ones(n_dists) / n_dists
        else:
            self._weights = np.array(weights) / np.sum(weights)

        if index is None:
            index = self._distributions[0][1].index

        if columns is None:
            columns = self._distributions[0][1].columns

        super().__init__(index=index, columns=columns)

    def _iloc(self, rowidx=None, colidx=None):
        dists = self._distributions
        weights = self.weights

        dists_subset = [(x[0], x[1].iloc[rowidx, colidx]) for x in dists]

        index_subset = dists_subset[0][1].index
        columns_subset = dists_subset[0][1].columns

        return Mixture(
            distributions=dists_subset,
            weights=weights,
            index=index_subset,
            columns=columns_subset,
        )

    def _iat(self, rowidx=None, colidx=None):
        dists = self._distributions
        weights = self.weights

        dists_subset = [(x[0], x[1].iat[rowidx, colidx]) for x in dists]

        return Mixture(distributions=dists_subset, weights=weights)

    def _mean(self):
        r"""Return expected value of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns the expectation :math:`\mathbb{E}[X]`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        expected value of distribution (entry-wise)
        """
        return self._average("mean")

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns :math:`\mathbb{V}[X] = \mathbb{E}\left(X - \mathbb{E}[X]\right)^2`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        variance of distribution (entry-wise)
        """
        weights = self._weights
        var_mean = self._average("var")
        mixture_mean = self._average("mean")

        means = [d.mean() for _, d in self._distributions]
        mean_var = [(m - mixture_mean) ** 2 for m in means]

        if self.ndim > 0:
            var_mean_var = self._average_df(mean_var, weights=weights)
        else:
            var_mean_var = np.average(mean_var, weights=weights)

        return var_mean + var_mean_var

    def _average(self, method, x=None, weights=None):
        """Average a method over the mixture components."""
        if x is None:
            args = ()
        else:
            args = (x,)

        vals = [getattr(d, method)(*args) for _, d in self._distributions]

        if self.ndim > 0:
            return self._average_df(vals, weights=weights)
        else:
            return np.average(vals, weights=weights)

    def _average_df(self, df_list, weights=None):
        """Average a list of `pd.DataFrame` objects, with weights."""
        if weights is None and hasattr(self, "_weights"):
            weights = self._weights
        elif weights is None:
            weights = np.ones(len(df_list)) / len(df_list)

        n_df = len(df_list)
        df_weighted = [df * w for df, w in zip(df_list, weights)]
        df_concat = pd.concat(df_weighted, axis=1, keys=range(n_df))
        df_res = df_concat.T.groupby(level=-1).sum().T
        return df_res

    def _pdf(self, x):
        """Probability density function."""
        return self._average("pdf", x)

    def _cdf(self, x):
        """Cumulative distribution function."""
        return self._average("cdf", x)

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        indep_rows = self.indep_rows
        indep_cols = self.indep_cols

        if n_samples is None:
            N = 1
        else:
            N = n_samples

        # deal with fully dependent case
        if not indep_rows and not indep_cols or self.ndim == 0:
            return self._sample_blocked(n_samples=n_samples)

        # we know that indep_rows and indep_cols are True, and self.ndim > 0
        rd_size = list(self.shape)
        full_size = list(self.shape)
        rd_size[0] *= N
        full_size[0] *= N

        if indep_rows:
            rd_size[0] = 1
        if indep_cols:
            rd_size[1] = 1

        n_dist = len(self._distributions)
        selector = np.random.choice(n_dist, size=rd_size, p=self._weights)
        indicators = [selector == i for i in range(n_dist)]
        indicators = [np.broadcast_to(ind, full_size) for ind in indicators]

        dists = [d[1] for d in self._distributions]
        raw_samples = [d.sample(N).values for d in dists]
        masked_samples = [ind * raw for ind, raw in zip(indicators, raw_samples)]
        masked_samples = np.array(masked_samples)
        sample = masked_samples.sum(axis=0)

        if n_samples is None:
            spl_index = self.index
        else:
            spl_index = pd.MultiIndex.from_product([range(N), self.index])

        spl = pd.DataFrame(sample, index=spl_index, columns=self.columns)
        return spl

    def _sample_blocked(self, n_samples):
        """Sample from the distribution with blocked rows and columns."""
        if n_samples is None:
            N = 1
        else:
            N = n_samples

        n_dist = len(self._distributions)

        selector = np.random.choice(n_dist, size=N, p=self._weights)

        samples = [self._distributions[i][1].sample() for i in selector]

        if n_samples is None:
            return samples[0]
        elif self.ndim > 0:
            return pd.concat(samples, axis=0, keys=range(N))
        else:  # if self.ndim == 0 and n_samples is not None
            return pd.DataFrame(samples, columns=self.columns)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        from skpro.distributions.normal import Normal

        # 2D array case
        index = pd.RangeIndex(3)
        columns = pd.Index(["a", "b"])
        normal1 = Normal(mu=0, sigma=1, index=index, columns=columns)
        normal2 = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, columns=columns)

        dists = [("normal1", normal1), ("normal2", normal2)]
        dists2 = [normal1, normal2]  # to check case without names

        params1 = {"distributions": dists}
        params2 = {"distributions": dists2, "weights": [0.3, 0.7]}

        # scalar case
        normal3 = Normal(mu=0, sigma=1)
        normal4 = Normal(mu=3, sigma=2)
        dists = [("normal3", normal3), ("normal4", normal4)]
        dists2 = [normal3, normal4]

        params3 = {"distributions": dists2}
        params4 = {"distributions": dists, "weights": [0.3, 0.7]}

        # more than 2 distributions
        normal5 = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=2, columns=columns)
        normal6 = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=0.5, columns=columns)
        dists3 = [normal1, normal2, normal5, normal6]
        params5 = {"distributions": dists3}

        return [params1, params2, params3, params4, params5]



"""Example generation for testing.

Exports examples of in-memory data containers, useful for testing as fixtures.

Examples come in clusters, tagged by scitype: str, index: int, and metadata: dict.

All examples with the same index are considered "content-wise the same", i.e.,
representing the same abstract data object. They differ by mtype, i.e.,
machine type, which is the specific in-memory representation.

If an example returns None, it indicates that representation
with that specific mtype is not possible.

If the tag "lossy" is True, the representation is considered incomplete,
e.g., metadata such as column names are missing.

Types of tests that can be performed with these examples:

* the mtype and scitype of the example should be correctly inferred by checkers.
* the metadata of hte example should be correctly inferred by checkers.
* conversions from non-lossy representations to any other ones
  should yield the element exactly, identically, for examples of the same index.
"""

import numpy as np
import pandas as pd

from skpro.datatypes._base import BaseExample

###
# example 0: univariate


class _UnivTable(BaseExample):
    _tags = {
        "scitype": "Table",
        "index": 0,
        "metadata": {
            "is_univariate": True,
            "is_empty": False,
            "has_nans": False,
            "n_instances": 4,
            "n_features": 1,
            "feature_names": ["a"],
        },
    }


class _UnivTableDf(_UnivTable):
    _tags = {
        "mtype": "pd_DataFrame_Table",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        return pd.DataFrame({"a": [1, 4, 0.5, -3]})


class _UnivTableNumpy2D(_UnivTable):
    _tags = {
        "mtype": "numpy2D",
        "python_dependencies": None,
        "lossy": True,
    }

    def build(self):
        return np.array([[1], [4], [0.5], [-3]])


class _UnivTableNumpy1D(_UnivTable):
    _tags = {
        "mtype": "numpy1D",
        "python_dependencies": None,
        "lossy": True,
    }

    def build(self):
        return np.array([1, 4, 0.5, -3])


class _UnivTableSeries(_UnivTable):
    _tags = {
        "mtype": "pd_Series_Table",
        "python_dependencies": None,
        "lossy": True,
    }

    def build(self):
        return pd.Series([1, 4, 0.5, -3])


class _UnivTableListOfDict(_UnivTable):
    _tags = {
        "mtype": "list_of_dict",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        return [{"a": 1.0}, {"a": 4.0}, {"a": 0.5}, {"a": -3.0}]


class _UnivTablePolarsEager(_UnivTable):
    _tags = {
        "mtype": "polars_eager_table",
        "python_dependencies": ["polars", "pyarrow"],
        "lossy": False,
    }

    def build(self):
        import polars as pl

        df = pd.DataFrame({"a": [1, 4, 0.5, -3]})
        return pl.DataFrame(df)


class _UnivTablePolarsLazy(_UnivTable):
    _tags = {
        "mtype": "polars_lazy_table",
        "python_dependencies": ["polars", "pyarrow"],
        "lossy": False,
    }

    def build(self):
        import polars as pl

        df = pd.DataFrame({"a": [1, 4, 0.5, -3]})
        return pl.LazyFrame(df)


###
# example 1: multivariate


class _MultivTable(BaseExample):
    _tags = {
        "scitype": "Table",
        "index": 1,
        "metadata": {
            "is_univariate": False,
            "is_empty": False,
            "has_nans": False,
            "n_instances": 4,
            "n_features": 2,
            "feature_names": ["a", "b"],
        },
    }


class _MultivTableDf(_MultivTable):
    _tags = {
        "mtype": "pd_DataFrame_Table",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        return pd.DataFrame({"a": [1, 4, 0.5, -3], "b": [3, 7, 2, -3 / 7]})


class _MultivTableNumpy2D(_MultivTable):
    _tags = {
        "mtype": "numpy2D",
        "python_dependencies": None,
        "lossy": True,
    }

    def build(self):
        return np.array([[1, 3], [4, 7], [0.5, 2], [-3, -3 / 7]])


class _MultivTableNumpy1D(_MultivTable):
    _tags = {
        "mtype": "numpy1D",
        "python_dependencies": None,
        "lossy": None,
    }

    def build(self):
        return None


class _MultivTableSeries(_MultivTable):
    _tags = {
        "mtype": "pd_Series_Table",
        "python_dependencies": None,
        "lossy": None,
    }

    def build(self):
        return None


class _MultivTableListOfDict(_MultivTable):
    _tags = {
        "mtype": "list_of_dict",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        return [
            {"a": 1.0, "b": 3.0},
            {"a": 4.0, "b": 7.0},
            {"a": 0.5, "b": 2.0},
            {"a": -3.0, "b": -3 / 7},
        ]


class _MultivTablePolarsEager(_MultivTable):
    _tags = {
        "mtype": "polars_eager_table",
        "python_dependencies": ["polars", "pyarrow"],
        "lossy": False,
    }

    def build(self):
        import polars as pl

        df = pd.DataFrame({"a": [1, 4, 0.5, -3], "b": [3, 7, 2, -3 / 7]})
        return pl.DataFrame(df)


class _MultivTablePolarsLazy(_MultivTable):
    _tags = {
        "mtype": "polars_lazy_table",
        "python_dependencies": ["polars", "pyarrow"],
        "lossy": False,
    }

    def build(self):
        import polars as pl

        df = pd.DataFrame({"a": [1, 4, 0.5, -3], "b": [3, 7, 2, -3 / 7]})
        return pl.LazyFrame(df)



"""Interface adapter to statsmodels cox proportional hazards models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

import numpy as np

from skpro.survival.base import BaseSurvReg


class CoxPH(BaseSurvReg):
    """Cox proportional hazards model, partial likelihood or elastic net, statsmodels.

    Direct interface to ``statsmodels.duration.hazard_regression.PHReg``

    Implements vanilla partial likelihood minimization (``method="lpl"``),
    and elastic net regularized minimization (``method="elastic_net"``).

    Parameters
    ----------
    method : str, optional (default="lpl"), one of {"lpl", "elastic_net}
        Method used to fit the proportional hazards model.
        "lpl": log partial likelihood minimization, vanilla method.
        Corresponds to statsmodels PHReg.fit method.
        "elastic_net": elastic net regularization.
        Corresponds to statsmodels PHReg.fit_regularized method.
    ties : str, optional (default="breslow"), one of {"breslow", "efron"}
        Method used to deal with ties in the data.
        "breslow": Breslow's method.
        "efron": Efron's method.
    missing : str, optional (default="drop"), one of {"drop"}
        Method used to deal with missing data, statsmodels native.
        "drop": Drops all observations with missing data.
        Note: statsmodels currently supports only "drop" for missing data.
        For imputation, pipeline with sklearn.impute.SimpleImputer,
        or another sklearn composable imputer, via skpro Pipeline (or * dunder)
    strata : pd.Index element or coercible, optional (default=None)
        Strata to use for stratified estimation.
        loc column reference to X, to column to be used as strata (categorical)
        If None, no stratification is used.
        If not None, must be a single valid loc column index of X.
        Iterable is used in loc subsetting of X, i.e., X.loc[:, [strata]].
        Corresponds to statsmodels PHReg.strata parameter.
    alpha : float or iterable of float, optional (default=0.0)
        Used only if method="elastic_net", otherwise ignored.
        Regularization parameter for elastic net regularization.
        If float, regularization parameter is the same for all variables.
        If iterable, must be of same length as number of variables (columns) in X.
        Corresponds to statsmodels PHReg.fit_regularized alpha parameter.

    Attributes
    ----------
    results_: statsmodels PHRegResults instance
        results of the fitted model
    """

    _tags = {
        "capability:missing": False,
        "capability:survival": True,
        "python_dependencies": ["statsmodels"],
    }

    def __init__(
        self,
        method="lpl",
        ties="breslow",
        missing="drop",
        strata=None,
        alpha=0.0,
    ):
        self.method = method
        self.ties = ties
        self.missing = missing
        self.strata = strata
        self.alpha = alpha
        super().__init__()

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        self : reference to self
        """
        from statsmodels.duration.hazard_regression import PHReg

        self._y_cols = y.columns

        endog = y.to_numpy().flatten()
        exog = X
        status = 1 - C.to_numpy().flatten() if C is not None else None

        params = {
            "ties": self.ties,
            "missing": self.missing,
        }

        if self.strata is not None:
            exog, strata = self._get_strata(exog)
            params["strata"] = strata

        model = PHReg(endog=endog, exog=exog, status=status, **params)
        self.model_ = model

        # fit model
        if self.method == "lpl":
            self.results_ = model.fit()
        else:
            self.results_ = model.fit_regularized(method=self.method, alpha=self.alpha)

        return self

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        from copy import deepcopy

        from skpro.distributions.adapters.statsmodels import empirical_from_rvdf

        # boilerplate code to create correct output index
        index = X.index
        y_cols = self._y_cols  # columns from y in fit, not automatically stored
        columns = y_cols

        # get results from statsmodels
        results = self.results_
        params = results.params
        exog = X.to_numpy()
        n_exog = exog.shape[0]

        # unreported bug in statsmodels PHReg.get_distribution
        # stratum_rows is not set correctly, uses rows from fit
        # we set it manually here
        model = self.model_
        model = deepcopy(model)
        model.surv.stratum_rows = [np.array([ix for ix in range(n_exog)])]

        if self.strata is not None:
            exog, strata = self._get_strata(exog)
            kwargs = {"params": params, "exog": exog, "strata": strata}
        else:
            kwargs = {"params": params, "exog": exog}

        # produce predictions from statsmodels
        # contrary to the documentation, this returns rv_discrete_float
        dist = model.get_distribution(**kwargs)
        dist.xk = dist.xk[:n_exog]
        dist.pk = dist.pk[:n_exog]

        # convert results to skpro BaseDistribution child instance
        y_pred = empirical_from_rvdf(dist=dist, index=index, columns=columns)
        return y_pred

    def _get_strata(self, X):
        """Get strata from X.

        Parameters
        ----------
        X : pandas DataFrame
            feature data frame to get strata from

        Returns
        -------
        X_wo_strata : pandas DataFrame
            X without strata column
        strata : 1D np.ndarray
            strata column from X, coerced to 1D np.ndarray
        """
        strata = X.loc[:, [self.strata]].to_numpy().flatten()
        X_wo_strata = X.drop(columns=[self.strata])
        return X_wo_strata, strata

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "method": "elastic_net",
            "alpha": 0.1,
            "ties": "efron",
        }

        return [params1, params2]



# LEGACY MODULE - TODO: remove or refactor

from .base import Controller, View


class RawView(View):
    """Raw view"""

    def parse(self, data):
        return str(data)


class ItemView(View):
    """Item view"""

    def __init__(self, key):
        self.key = key

    def parse(self, data):
        if self.key in data:
            return str(data[self.key])
        else:
            return self.key


class InfoView(View):
    """InfoView"""

    def __init__(self, key="description", with_group=False):
        self.key = key
        self.with_group = with_group

    def parse(self, data):
        if self.with_group:
            return "%s [%s]" % (data[self.key], data["group"])

        return data[self.key]


class InfoController(Controller):
    """InfoController"""

    def identifier(self):
        return "InfoController()"

    def __repr__(self):
        return "InfoController()"

    def __str__(self):
        return "Info"

    def run(self, model):
        return {
            "description": str(model).replace("\n           ", ""),
            "repr": repr(model),
            "group": getattr(model, "group", "None"),
        }



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Base class for data types."""

__author__ = ["fkiraly"]

from skpro.datatypes._base import BaseDatatype


class BaseTable(BaseDatatype):
    """Base class for Table data types.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": None,  # any string
        "name_python": None,  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": None,
    }

    def __init__(
        self,
        is_univariate=None,
        is_empty=None,
        has_nans=None,
        n_instances=None,
        n_features=None,
        feature_names=None,
    ):
        self.is_univariate = is_univariate
        self.is_empty = is_empty
        self.has_nans = has_nans
        self.n_instances = n_instances
        self.n_features = n_features
        self.feature_names = feature_names

        super().__init__()



"""Bayesian Conjugate Linear Regression Estimator."""

__author__ = ["meraldoantonio"]

import numpy as np
import pandas as pd

from skpro.distributions import Normal
from skpro.regression.base import BaseProbaRegressor


class BayesianConjugateLinearRegressor(BaseProbaRegressor):
    """
    Bayesian probabilistic estimator for linear regression.

    This estimator models the relationship between features `X` and target `t` using
    a Bayesian linear regression framework with conjugate priors.
    Specifically, the prior and posterior of the coefficients (`w`) are
    both  multivariate Gaussians.

    Model Assumptions
    -----------------
    - **Prior for coefficients (`w`)**:
      Prior for coefficients `w` (including intercept) follow a multivariate Gaussian:
      ```
      w ~ N(coefs_prior_mu, coefs_prior_cov)
      ```
      where:
      - `coefs_prior_mu`: Mean vector of the prior distribution for `w`.
      - `coefs_prior_cov`: Covariance matrix of the prior distribution for `w`.

    - **Likelihood**:
      The likelihood of target `t` given features `X` and coefficients `w` is Gaussian:
      ```
      t | X, w ~ N(X @ w, sigma^2)
      ```
      where:
      - `sigma^2 = 1 / noise_precision`: Variance of the noise in the data.
      - `noise_precision`: Known precision (inverse variance) of the noise.

    - **Posterior for coefficients (`w`)**:
      Using Bayesian conjugacy, posterior of `w` remains Gaussian:
      ```
      w | X, t ~ N(coefs_posterior_mu, coefs_posterior_cov)
      ```
      with:
      ```
      coefs_posterior_cov = (coefs_prior_precision + noise_precision * X.T @ X)^(-1)
      coefs_posterior_mu = coefs_posterior_cov @ (
          coefs_prior_precision @ coefs_prior_mu + noise_precision * X.T @ y
      )
      ```
      where:
      - `coefs_prior_precision = inv(coefs_prior_cov)`.

    - **Predictive distribution**:
      For a new observation `x`, the predictive distribution of `y` is:
      ```
      y_pred | x ~ N(x.T @ coefs_posterior_mu, pred_var)
      ```
      where:
      ```
      pred_var = x.T @ coefs_posterior_cov @ x + 1 / noise_precision
      ```

    Example
    -------
    >>> from skpro.regression.bayesian.bayesian_conjugate import (
    ...     BayesianConjugateLinearRegressor,
    ... )  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>> from sklearn.model_selection import train_test_split  # doctest: +SKIP

    >>> # Load dataset
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)  # doctest: +SKIP

    >>> # Center the training data
    >>> X_train -= X_train.mean(axis=0)  # doctest: +SKIP

    >>> # Initialize model
    >>> bayes_model = BayesianConjugateLinearRegressor(
    ...     coefs_prior_mu=None,
    ...     coefs_prior_cov=np.eye(X_train.shape[1]),
    ...     noise_precision=1.0,
    ... )  # doctest: +SKIP

    >>> # Fit the model
    >>> bayes_model.fit(X_train, y_train)  # doctest: +SKIP

    >>> # Predict probabilities (returns an skpro Normal distribution)
    >>> y_test_pred_proba = bayes_model.predict_proba(X_test)  # doctest: +SKIP

    >>> # Predict point estimates (mean of the predicted distribution)
    >>> y_test_pred = bayes_model.predict(X_test)  # doctest: +SKIP
    """

    _tags = {
        "authors": ["meraldoantonio"],
        "capability:multioutput": False,
        "capability:missing": True,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
    }

    def __init__(self, coefs_prior_cov, coefs_prior_mu=None, noise_precision=1):
        """Initialize regressor by providing coefficent priors and noise precision.

        Parameters
        ----------
        coefs_prior_cov : 2D np.ndarray, required
            Covariance matrix of the prior for intercept and coefficients.
            If list of lists, will be converted to a 2D np.array.
            Must be positive-definite.

        coefs_prior_mu : np.ndarray column vector, optional
            Mean vector of the prior for intercept and coefficients.
            The zeroth element of the vector is the prior for the intercept.
            If not provided, assumed to be a column vector of zeroes.

        noise_precision : float
            Known precision of the Gaussian likelihood noise (beta)
            This is the inverse of the noise variance.
        """
        if coefs_prior_cov is None:
            raise ValueError("`coefs_prior_cov` must be provided.")
        else:
            self.coefs_prior_cov = coefs_prior_cov

        # Set coefs_prior_mu to a zero vector if not provided
        if coefs_prior_mu is None:
            self.coefs_prior_mu = np.zeros((self.coefs_prior_cov.shape[0], 1))
        # Ensure coefs_prior_mu is a column vector
        elif coefs_prior_mu.ndim != 2 or coefs_prior_mu.shape[1] != 1:
            raise ValueError(
                "coefs_prior_mu must be a column vector with shape (n_features, 1)."
            )
        else:
            self.coefs_prior_mu = coefs_prior_mu

        # Validate dimensions of coefs_prior_mu and coefs_prior_cov
        if self.coefs_prior_mu.shape[0] != self.coefs_prior_cov.shape[0]:
            raise ValueError(
                "Dimensionality of `coefs_prior_mu` and `coefs_prior_cov` must match."
            )

        self.noise_precision = noise_precision

        super().__init__()

    def _fit(self, X, y):
        """Fit the Bayesian Linear Regressor to the observed data.

        Parameters
        ----------
        X : pandas DataFrame
            Feature matrix (n_samples, n_features).
        y : pandas Series
            Target vector (n_samples,).

        Returns
        -------
        self : reference to self
        """
        self._y_cols = y.columns

        # Construct the prior mean and covariance
        self._coefs_prior_mu = self.coefs_prior_mu
        self._coefs_prior_cov = self.coefs_prior_cov
        self._coefs_prior_precision = np.linalg.inv(self._coefs_prior_cov)

        # Perform Bayesian inference
        (
            self._coefs_posterior_mu,
            self._coefs_posterior_cov,
        ) = self._perform_bayesian_inference(
            X, y, self._coefs_prior_mu, self._coefs_prior_cov
        )
        return self

    def _predict_proba(self, X):
        """Predict the distribution over outputs for given features.

        Parameters
        ----------
        X : pandas DataFrame
            Feature matrix (n_samples, n_features).

        Returns
        -------
        y_pred : Normal
            Predicted Normal distribution for outputs.
        """
        idx = X.index
        if isinstance(X, pd.DataFrame):
            X = X.values

        # Predictive mean: X * posterior_mu
        pred_mu = X @ self._coefs_posterior_mu

        # Compute predictive variance for each data point
        pred_var_all_x_i = []
        for i in range(X.shape[0]):
            x_i = X[i, :].reshape(1, -1)
            pred_var_x_i = (
                x_i @ self._coefs_posterior_cov @ x_i.T + 1 / self.noise_precision
            )
            pred_var_all_x_i.append(pred_var_x_i.item())

        pred_var_all_x_i = np.array(pred_var_all_x_i)
        pred_sigma = np.sqrt(pred_var_all_x_i)  # Compute standard deviation

        self._pred_mu = pred_mu
        self._pred_sigma = pred_sigma
        self._pred_var = pred_var_all_x_i

        mus = pred_mu.reshape(-1, 1).tolist()  # Convert to list of lists
        sigmas = pred_sigma.reshape(-1, 1).tolist()  # Convert to list of lists

        # Return results as skpro Normal distribution
        self._y_pred = Normal(
            mu=mus,
            sigma=sigmas,
            columns=self._y_cols,
            index=idx,
        )
        return self._y_pred

    def _perform_bayesian_inference(self, X, y, coefs_prior_mu, coefs_prior_cov):
        """Perform Bayesian inference for linear regression.

        Obtains the posterior distribution using normal conjugacy formula.

        Parameters
        ----------
        X : pandas DataFrame
            Feature matrix (n_samples, n_features).
        y : pandas Series
            Observed target vector (n_samples,).
        coefs_prior_mu : np.ndarray
            Mean vector of the prior Normal distribution for coefficients.
        coefs_prior_cov : np.ndarray
            Covariance matrix of the prior Normal distribution for coefficients.

        Returns
        -------
        coefs_posterior_mu : np.ndarray
            Mean vector of the posterior Normal distribution for coefficients.
        coefs_posterior_cov : np.ndarray
            Covariance matrix of the posterior Normal distribution for coefficients.
        """
        X = np.array(X)
        y = np.array(y)

        # Compute prior precision from prior covariance
        coefs_prior_precision = np.linalg.inv(coefs_prior_cov)

        # Compute posterior precision and covariance
        coefs_posterior_precision = coefs_prior_precision + self.noise_precision * (
            X.T @ X
        )
        coefs_posterior_cov = np.linalg.inv(coefs_posterior_precision)
        coefs_posterior_mu = coefs_posterior_cov @ (
            coefs_prior_precision @ coefs_prior_mu + self.noise_precision * X.T @ y
        )

        return coefs_posterior_mu, coefs_posterior_cov

    def update(self, X, y):
        """Update the posterior with new data.

        Parameters
        ----------
        X : pandas DataFrame
            New feature matrix.
        y : pandas Series or DataFrame
            New target vector.

        Returns
        -------
        self : reference to self
        """
        # Ensure y is a DataFrame
        if isinstance(y, pd.Series):
            y = y.to_frame(name="y_train")

        (
            self._coefs_posterior_mu,
            self._coefs_posterior_cov,
        ) = self._perform_bayesian_inference(
            X, y, self._coefs_posterior_mu, self._coefs_posterior_cov
        )
        return self

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
        """
        params1 = {
            "coefs_prior_mu": np.zeros(10).reshape(-1, 1),
            "coefs_prior_cov": np.eye(10),
            "noise_precision": 1.0,
        }

        params2 = {
            "coefs_prior_mu": np.zeros(10).reshape(-1, 1),
            "coefs_prior_cov": np.eye(10),
            "noise_precision": 0.5,
        }

        return [params1, params2]



"""Cox proportional hazards models."""

from skpro.survival.coxph._coxnet_sksurv import CoxNet
from skpro.survival.coxph._coxph_lifelines import CoxPHlifelines
from skpro.survival.coxph._coxph_sksurv import CoxPHSkSurv
from skpro.survival.coxph._coxph_statsmodels import CoxPH

__all__ = ["CoxNet", "CoxPH", "CoxPHlifelines", "CoxPHSkSurv"]



#!/usr/bin/env python3 -u
"""Utilities for pandas adapbation."""

__author__ = ["fkiraly"]


def df_map(x):
    """Access map or applymap, of DataFrame.

    In pandas 2.1.0, applymap was deprecated in favor of the newly introduced map.
    To ensure compatibility with older versions, we use map if available,
    otherwise applymap.

    Parameters
    ----------
    x : assumed pd.DataFrame

    Returns
    -------
    x.map, if available, otherwise x.applymap
        Note: returns method itself, not result of method call
    """
    if hasattr(x, "map"):
        return x.map
    else:
        return x.applymap



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Empirical distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd


def empirical_from_rvdf(dist, index=None, columns=None):
    """Convert a statsmodels rv_discrete_float to an skpro Empirical object.

    Parameters
    ----------
    dist : rv_discrte_float object
        Instance of rv_discrete.
    index : pd.Index or coercible, optional
        Index of the resulting empirical distribution.
        Must be the same length as dist.
    columns : pd.Index or coercible, optional
        Columns of the resulting empirical distribution.
        Must be of length 1.
    """
    from skpro.distributions.empirical import Empirical

    if index is None:
        index = pd.RangeIndex(len(dist))

    xk = dist.xk
    pk = dist.pk

    xks = [xk[i] for i in range(len(xk))]
    pks = [pk[i] for i in range(len(pk))]

    lens = [len(xk) for xk in xks]
    idxs_inst = [np.repeat(index[i], leni) for i, leni in enumerate(lens)]
    idx_inst_flat = np.concatenate(idxs_inst)
    idx_spl = [np.arange(leni) for leni in lens]
    idx_spl_flat = np.concatenate(idx_spl)

    idx_mult = pd.MultiIndex.from_arrays([idx_spl_flat, idx_inst_flat])

    spl = pd.DataFrame(np.concatenate(xks), index=idx_mult, columns=columns)
    weights = pd.Series(np.concatenate(pks), index=idx_mult)

    emp = Empirical(
        spl=spl, weights=weights, time_indep=True, index=index, columns=columns
    )
    return emp



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Tuning of probabilistic supervised regressors."""

__author__ = ["fkiraly"]
__all__ = ["GridSearchCV", "RandomizedSearchCV"]

import numpy as np
import pandas as pd
from sklearn.model_selection import ParameterGrid, ParameterSampler, check_cv

from skpro.benchmarking.evaluate import evaluate
from skpro.regression.base._delegate import _DelegatedProbaRegressor
from skpro.utils.parallel import parallelize


class BaseGridSearch(_DelegatedProbaRegressor):
    _tags = {
        "estimator_type": "regressor",
        "capability:multioutput": True,
        "capability:missing": True,
    }

    def __init__(
        self,
        estimator,
        cv,
        backend="loky",
        refit=True,
        scoring=None,
        verbose=0,
        return_n_best_estimators=1,
        error_score=np.nan,
        backend_params=None,
    ):
        self.estimator = estimator
        self.cv = cv
        self.backend = backend
        self.refit = refit
        self.scoring = scoring
        self.verbose = verbose
        self.return_n_best_estimators = return_n_best_estimators
        self.error_score = error_score
        self.backend_params = backend_params

        super().__init__()

        tags_to_clone = [
            "capability:multioutput",
            "capability:missing",
            "capability:survival",
        ]
        self.clone_tags(estimator, tags_to_clone)

    # attribute for _DelegatedProbaRegressor, which then delegates
    #     all non-overridden methods are same as of getattr(self, _delegate_name)
    #     see further details in _DelegatedProbaRegressor docstring
    _delegate_name = "best_estimator_"

    def _get_fitted_params(self):
        """Get fitted parameters.

        Returns
        -------
        fitted_params : dict
            A dict containing the best hyper parameters and the parameters of
            the best estimator (if available), merged together with the former
            taking precedence.
        """
        fitted_params = {}
        try:
            fitted_params = self.best_estimator_.get_fitted_params()
        except NotImplementedError:
            pass
        fitted_params = {**fitted_params, **self.best_params_}
        fitted_params.update(self._get_fitted_params_default())

        return fitted_params

    def _run_search(self, evaluate_candidates):
        raise NotImplementedError("abstract method")

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        cv = check_cv(self.cv)

        # scoring = check_scoring(self.scoring, obj=self)
        scoring = self.scoring
        scoring_name = f"test_{scoring.name}"

        backend = self.backend
        backend_params = self.backend_params if self.backend_params else {}

        def _fit_and_score(params, meta):
            # Clone estimator.
            estimator = self.estimator.clone()

            # Set parameters.
            estimator.set_params(**params)

            # Evaluate.
            out = evaluate(
                estimator,
                cv=cv,
                X=X,
                y=y,
                C=C,
                scoring=scoring,
                error_score=self.error_score,
            )

            # Filter columns.
            out = out.filter(items=[scoring_name, "fit_time", "pred_time"], axis=1)

            # Aggregate results.
            out = out.mean()
            out = out.add_prefix("mean_")

            # Add parameters to output table.
            out["params"] = params

            return out

        def evaluate_candidates(candidate_params):
            candidate_params = list(candidate_params)

            if self.verbose > 0:
                n_candidates = len(candidate_params)
                n_splits = cv.get_n_splits(y)
                print(  # noqa
                    "Fitting {} folds for each of {} candidates,"
                    " totalling {} fits".format(
                        n_splits, n_candidates, n_candidates * n_splits
                    )
                )

            out = parallelize(
                fun=_fit_and_score,
                iter=candidate_params,
                backend=backend,
                backend_params=backend_params,
            )

            if len(out) < 1:
                raise ValueError(
                    "No fits were performed. "
                    "Was the CV iterator empty? "
                    "Were there no candidates?"
                )

            return out

        # Run grid-search cross-validation.
        results = self._run_search(evaluate_candidates)

        results = pd.DataFrame(results)

        # Rank results, according to whether greater is better for the given scoring.
        results[f"rank_{scoring_name}"] = results.loc[:, f"mean_{scoring_name}"].rank(
            ascending=scoring.get_tag("lower_is_better")
        )

        self.cv_results_ = results

        # Select best parameters.
        self.best_index_ = results.loc[:, f"rank_{scoring_name}"].argmin()
        # Raise error if all fits in evaluate failed because all score values are NaN.
        if self.best_index_ == -1:
            raise RuntimeError(
                f"""All fits of estimator failed,
                set error_score='raise' to see the exceptions.
                Failed estimator: {self.estimator}"""
            )
        self.best_score_ = results.loc[self.best_index_, f"mean_{scoring_name}"]
        self.best_params_ = results.loc[self.best_index_, "params"]
        self.best_estimator_ = self.estimator.clone().set_params(**self.best_params_)

        # Refit model with best parameters.
        if self.refit:
            self.best_estimator_.fit(X, y, C=C)

        # Sort values according to rank
        results = results.sort_values(
            by=f"rank_{scoring_name}",
            ascending=True,
        )
        # Select n best estimator
        self.n_best_estimators_ = []
        self.n_best_scores_ = []
        for i in range(self.return_n_best_estimators):
            params = results["params"].iloc[i]
            rank = results[f"rank_{scoring_name}"].iloc[i]
            rank = str(int(rank))
            estimator = self.estimator.clone().set_params(**params)
            # Refit model with best parameters.
            if self.refit:
                estimator.fit(X, y)
            self.n_best_estimators_.append((rank, estimator))
            # Save score
            score = results[f"mean_{scoring_name}"].iloc[i]
            self.n_best_scores_.append(score)

        return self

    def _get_delegate(self):
        if self.is_fitted and not self.refit:
            raise RuntimeError(
                f"In {self.__class__.__name__}, refit must be True to make predictions,"
                f" but found refit=False. If refit=False, {self.__class__.__name__} can"
                " be used only to tune hyper-parameters, as a parameter estimator."
            )
        return getattr(self, self._delegate_name)


class GridSearchCV(BaseGridSearch):
    """Perform grid-search cross-validation to find optimal model parameters.

    The estimator is fit on the initial window and then temporal
    cross-validation is used to find the optimal parameter.

    Grid-search cross-validation is performed based on a cross-validation
    iterator encoding the cross-validation scheme, the parameter grid to
    search over, and (optionally) the evaluation metric for comparing model
    performance. As in scikit-learn, tuning works through the common
    hyper-parameter interface which allows to repeatedly fit and evaluate
    the same estimator with different hyper-parameters.

    Parameters
    ----------
    estimator : estimator object
        The estimator should implement the skpro estimator
        interface. Either the estimator must contain a "score" function,
        or a scoring function must be passed.

    cv : cross-validation generator or an iterable
        e.g. KFold(n_splits=3)

    param_grid : dict or list of dictionaries
        Model tuning parameters of the estimator to evaluate

    scoring : skpro metric (BaseMetric), str, or callable, optional (default=None)
        scoring metric to use in tuning the estimator

        * skpro metric objects (BaseMetric) descendants can be searched
        with the ``registry.all_estimators`` search utility,
        for instance via ``all_estimators("metric", as_dataframe=True)``

        * If callable, must have signature
        `(y_true: pd.DataFrame, y_pred: BaseDistribution) -> float`,
        assuming y_true, y_pred are of the same length, lower being better,
        Metrics in skpro.metrics are all of this form.

        * If str, uses registry.resolve_alias to resolve to one of the above.
          Valid strings are valid registry.craft specs, which include
          string repr-s of any BaseMetric object, e.g., "MeanSquaredError()";
          and keys of registry.ALIAS_DICT referring to metrics.

        * If None, defaults to CRPS()

    refit : bool, optional (default=True)
        True = refit the estimator with the best parameters on the entire data in fit
        False = no refitting takes place. The estimator cannot be used to predict.
        This is to be used to tune the hyperparameters, and then use the estimator
        as a parameter estimator, e.g., via get_fitted_params or PluginParamsestimator.

    verbose: int, optional (default=0)

    return_n_best_estimators : int, default=1
        In case the n best estimator should be returned, this value can be set
        and the n best estimators will be assigned to n_best_estimators_

    error_score : numeric value or the str 'raise', optional (default=np.nan)
        The test score returned when a estimator fails to be fitted.

    return_train_score : bool, optional (default=False)

    backend : {"dask", "loky", "multiprocessing", "threading"}, by default "loky".
        Runs parallel evaluate if specified and `strategy` is set as "refit".

        - "None": executes loop sequentally, simple list comprehension
        - "loky", "multiprocessing" and "threading": uses ``joblib.Parallel`` loops
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``
        - "dask": uses ``dask``, requires ``dask`` package in environment

        Recommendation: Use "dask" or "loky" for parallel evaluate.
        "threading" is unlikely to see speed ups due to the GIL and the serialization
        backend (``cloudpickle``) for "dask" and "loky" is generally more robust
        than the standard ``pickle`` library used in "multiprocessing".

    error_score : "raise" or numeric, default=np.nan
        Value to assign to the score if an exception occurs in estimator fitting. If set
        to "raise", the exception is raised. If a numeric value is given,
        FitFailedWarning is raised.

    backend_params : dict, optional
        additional parameters passed to the backend as config.
        Directly passed to ``utils.parallel.parallelize``.
        Valid keys depend on the value of ``backend``:

        - "None": no additional parameters, ``backend_params`` is ignored
        - "loky", "multiprocessing" and "threading": default ``joblib`` backends
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          with the exception of ``backend`` which is directly controlled by ``backend``.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``.
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          ``backend`` must be passed as a key of ``backend_params`` in this case.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "dask": any valid keys for ``dask.compute`` can be passed, e.g., ``scheduler``

    Attributes
    ----------
    best_index_ : int
    best_score_: float
        Score of the best model
    best_params_ : dict
        Best parameter values across the parameter grid
    best_estimator_ : estimator
        Fitted estimator with the best parameters
    cv_results_ : dict
        Results from grid search cross validation
    n_splits_: int
        Number of splits in the data for cross validation
    refit_time_ : float
        Time (seconds) to refit the best estimator
    scorer_ : function
        Function used to score model
    n_best_estimators_: list of tuples ("rank", <estimator>)
        The "rank" is in relation to best_estimator_
    n_best_scores_: list of float
        The scores of n_best_estimators_ sorted from best to worst
        score of estimators

    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.model_selection import KFold, ShuffleSplit, train_test_split

    >>> from skpro.metrics import CRPS
    >>> from skpro.model_selection import GridSearchCV
    >>> from skpro.regression.residual import ResidualDouble

    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> y = pd.DataFrame(y)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)

    >>> cv = KFold(n_splits=3)

    >>> estimator = ResidualDouble(LinearRegression())
    >>> param_grid = {"estimator__fit_intercept" : [True, False]}
    >>> gscv = GridSearchCV(
    ...     estimator=estimator,
    ...     param_grid=param_grid,
    ...     cv=cv,
    ...     scoring=CRPS(),
    ... )
    >>> gscv.fit(X_train, y_train)
    GridSearchCV(...)
    >>> y_pred = gscv.predict(X_test)
    >>> y_pred_proba = gscv.predict_proba(X_test)
    """

    def __init__(
        self,
        estimator,
        cv,
        param_grid,
        scoring=None,
        refit=True,
        verbose=0,
        return_n_best_estimators=1,
        backend="loky",
        error_score=np.nan,
        backend_params=None,
    ):
        super().__init__(
            estimator=estimator,
            scoring=scoring,
            refit=refit,
            cv=cv,
            verbose=verbose,
            return_n_best_estimators=return_n_best_estimators,
            backend=backend,
            error_score=error_score,
            backend_params=backend_params,
        )
        self.param_grid = param_grid

    def _check_param_grid(self, param_grid):
        """_check_param_grid from sklearn 1.0.2, before it was removed."""
        if hasattr(param_grid, "items"):
            param_grid = [param_grid]

        for p in param_grid:
            for name, v in p.items():
                if isinstance(v, np.ndarray) and v.ndim > 1:
                    raise ValueError("Parameter array should be one-dimensional.")

                if isinstance(v, str) or not isinstance(v, (np.ndarray, list)):
                    raise ValueError(
                        "Parameter grid for parameter ({}) needs to"
                        " be a list or numpy array, but got ({})."
                        " Single values need to be wrapped in a list"
                        " with one element.".format(name, type(v))
                    )

                if len(v) == 0:
                    raise ValueError(
                        "Parameter values for parameter ({}) need "
                        "to be a non-empty sequence.".format(name)
                    )

    def _run_search(self, evaluate_candidates):
        """Search all candidates in param_grid."""
        self._check_param_grid(self.param_grid)
        return evaluate_candidates(ParameterGrid(self.param_grid))

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict
        """
        from sklearn.linear_model import LinearRegression
        from sklearn.model_selection import KFold

        from skpro.metrics import CRPS, PinballLoss
        from skpro.regression.residual import ResidualDouble
        from skpro.survival.compose._reduce_cond_unc import ConditionUncensored

        linreg1 = LinearRegression()
        linreg2 = LinearRegression(fit_intercept=False)

        param1 = {
            "estimator": ResidualDouble(LinearRegression()),
            "cv": KFold(n_splits=3),
            "param_grid": {"estimator": [linreg1, linreg2]},
            "scoring": CRPS(),
            "error_score": "raise",
        }

        param2 = {
            "estimator": ResidualDouble(LinearRegression()),
            "cv": KFold(n_splits=4),
            "param_grid": {"estimator__fit_intercept": [True, False]},
            "scoring": PinballLoss(),
            "error_score": "raise",
        }

        params3 = {
            "estimator": ConditionUncensored(ResidualDouble(LinearRegression())),
            "cv": KFold(n_splits=4),
            "param_grid": {"estimator__fit_intercept": [True, False]},
            "scoring": PinballLoss(),
            "error_score": "raise",
        }
        params = [param1, param2, params3]

        return params


class RandomizedSearchCV(BaseGridSearch):
    """Perform randomized-search cross-validation to find optimal model parameters.

    The estimator is fit on the initial window and then
    cross-validation is used to find the optimal parameter

    Randomized cross-validation is performed based on a cross-validation
    iterator encoding the cross-validation scheme, the parameter distributions to
    search over, and (optionally) the evaluation metric for comparing model
    performance. As in scikit-learn, tuning works through the common
    hyper-parameter interface which allows to repeatedly fit and evaluate
    the same estimator with different hyper-parameters.

    Parameters
    ----------
    estimator : estimator object
        The estimator should implement the skpro or scikit-learn estimator
        interface. Either the estimator must contain a "score" function,
        or a scoring function must be passed.

    cv : cross-validation generator or an iterable
        e.g. KFold(n_splits=3)

    param_distributions : dict or list of dicts
        Dictionary with parameters names (`str`) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.
        If a list of dicts is given, first a dict is sampled uniformly, and
        then a parameter is sampled using that dict as above.

    n_iter : int, default=10
        Number of parameter settings that are sampled. n_iter trades
        off runtime vs quality of the solution.

    scoring : skpro metric (BaseMetric), str, or callable, optional (default=None)
        scoring metric to use in tuning the estimator

        * skpro metric objects (BaseMetric) descendants can be searched
        with the ``registry.all_estimators`` search utility,
        for instance via ``all_estimators("metric", as_dataframe=True)``

        * If callable, must have signature
        `(y_true: pd.DataFrame, y_pred: BaseDistribution) -> float`,
        assuming y_true, y_pred are of the same length, lower being better,
        Metrics in skpro.metrics are all of this form.

        * If str, uses registry.resolve_alias to resolve to one of the above.
          Valid strings are valid registry.craft specs, which include
          string repr-s of any BaseMetric object, e.g., "MeanSquaredError()";
          and keys of registry.ALIAS_DICT referring to metrics.

        * If None, defaults to CRPS()

    refit : bool, optional (default=True)
        True = refit the estimator with the best parameters on the entire data in fit
        False = no refitting takes place. The estimator cannot be used to predict.
        This is to be used to tune the hyperparameters, and then use the estimator
        as a parameter estimator, e.g., via get_fitted_params or PluginParamsestimator.

    verbose : int, optional (default=0)

    return_n_best_estimators: int, default=1
        In case the n best estimator should be returned, this value can be set
        and the n best estimators will be assigned to n_best_estimators_

    random_state : int, RandomState instance or None, default=None
        Pseudo random number generator state used for random uniform sampling
        from lists of possible values instead of scipy.stats distributions.
        Pass an int for reproducible output across multiple
        function calls.

    backend : {"dask", "loky", "multiprocessing", "threading"}, by default "loky".
        Runs parallel evaluate if specified and `strategy` is set as "refit".

        - "None": executes loop sequentally, simple list comprehension
        - "loky", "multiprocessing" and "threading": uses ``joblib.Parallel`` loops
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``
        - "dask": uses ``dask``, requires ``dask`` package in environment

        Recommendation: Use "dask" or "loky" for parallel evaluate.
        "threading" is unlikely to see speed ups due to the GIL and the serialization
        backend (``cloudpickle``) for "dask" and "loky" is generally more robust
        than the standard ``pickle`` library used in "multiprocessing".

    error_score : "raise" or numeric, default=np.nan
        Value to assign to the score if an exception occurs in estimator fitting. If set
        to "raise", the exception is raised. If a numeric value is given,
        FitFailedWarning is raised.

    backend_params : dict, optional
        additional parameters passed to the backend as config.
        Directly passed to ``utils.parallel.parallelize``.
        Valid keys depend on the value of ``backend``:

        - "None": no additional parameters, ``backend_params`` is ignored
        - "loky", "multiprocessing" and "threading": default ``joblib`` backends
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          with the exception of ``backend`` which is directly controlled by ``backend``.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``.
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          ``backend`` must be passed as a key of ``backend_params`` in this case.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "dask": any valid keys for ``dask.compute`` can be passed, e.g., ``scheduler``

    Attributes
    ----------
    best_index_ : int
    best_score_: float
        Score of the best model
    best_params_ : dict
        Best parameter values across the parameter grid
    best_estimator_ : estimator
        Fitted estimator with the best parameters
    cv_results_ : dict
        Results from grid search cross validation
    n_best_estimators_: list of tuples ("rank", <estimator>)
        The "rank" is in relation to best_estimator_
    n_best_scores_: list of float
        The scores of n_best_estimators_ sorted from best to worst
        score of estimators

    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.model_selection import KFold, ShuffleSplit, train_test_split

    >>> from skpro.metrics import CRPS
    >>> from skpro.model_selection import RandomizedSearchCV
    >>> from skpro.regression.residual import ResidualDouble

    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> y = pd.DataFrame(y)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)

    >>> cv = KFold(n_splits=3)

    >>> estimator = ResidualDouble(LinearRegression())
    >>> param_distributions = {"estimator__fit_intercept" : [True, False]}
    >>> rscv = RandomizedSearchCV(
    ...     estimator=estimator,
    ...     param_distributions=param_distributions,
    ...     cv=cv,
    ...     scoring=CRPS(),
    ... )
    >>> rscv.fit(X_train, y_train)
    RandomizedSearchCV(...)
    >>> y_pred = rscv.predict(X_test)
    >>> y_pred_proba = rscv.predict_proba(X_test)
    """

    def __init__(
        self,
        estimator,
        cv,
        param_distributions,
        n_iter=10,
        scoring=None,
        refit=True,
        verbose=0,
        return_n_best_estimators=1,
        random_state=None,
        error_score=np.nan,
        backend_params=None,
    ):
        super().__init__(
            estimator=estimator,
            scoring=scoring,
            refit=refit,
            cv=cv,
            verbose=verbose,
            return_n_best_estimators=return_n_best_estimators,
            error_score=error_score,
            backend_params=backend_params,
        )
        self.param_distributions = param_distributions
        self.n_iter = n_iter
        self.random_state = random_state

    def _run_search(self, evaluate_candidates):
        """Search n_iter candidates from param_distributions."""
        return evaluate_candidates(
            ParameterSampler(
                self.param_distributions, self.n_iter, random_state=self.random_state
            )
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict
        """
        from sklearn.linear_model import LinearRegression
        from sklearn.model_selection import KFold

        from skpro.metrics import CRPS, PinballLoss
        from skpro.regression.residual import ResidualDouble
        from skpro.survival.compose._reduce_cond_unc import ConditionUncensored

        linreg1 = LinearRegression()
        linreg2 = LinearRegression(fit_intercept=False)

        param1 = {
            "estimator": ResidualDouble(LinearRegression()),
            "cv": KFold(n_splits=3),
            "param_distributions": {"estimator": [linreg1, linreg2]},
            "scoring": CRPS(),
            "error_score": "raise",
        }

        param2 = {
            "estimator": ResidualDouble(LinearRegression()),
            "cv": KFold(n_splits=4),
            "param_distributions": {"estimator__fit_intercept": [True, False]},
            "scoring": PinballLoss(),
            "error_score": "raise",
        }

        params3 = {
            "estimator": ConditionUncensored(ResidualDouble(LinearRegression())),
            "cv": KFold(n_splits=4),
            "param_distributions": {"estimator__fit_intercept": [True, False]},
            "scoring": PinballLoss(),
            "error_score": "raise",
        }
        params = [param1, param2, params3]

        return params



"""Test file for polars dataframes"""

import pandas as pd
import pytest
from skbase.utils.dependencies import _check_soft_dependencies
from sklearn.datasets import load_diabetes
from sklearn.model_selection import train_test_split

from skpro.tests.test_switch import run_test_module_changed

if _check_soft_dependencies(["polars", "pyarrow"], severity="none"):
    import polars as pl

    from skpro.datatypes import check_is_mtype, convert

TEST_ALPHAS = [0.05, 0.1, 0.25]


@pytest.fixture
def polars_load_diabetes_pandas():
    X, y = load_diabetes(return_X_y=True, as_frame=True)
    X = X.iloc[:75]
    y = y.iloc[:75]

    # typically y is returned as a pd.Series to we call y as a Dataframe here
    y = pd.DataFrame(y)

    X_train, X_test, y_train, _ = train_test_split(
        X, y, test_size=0.33, random_state=42
    )
    return [X_train, X_test, y_train]


@pytest.fixture
def estimator():
    from sklearn.linear_model import LinearRegression

    from skpro.regression.residual import ResidualDouble

    # refactor to use ResidualDouble with Linear Regression
    _estimator = ResidualDouble(LinearRegression())
    return _estimator


def _pd_to_pl(df):
    return convert(df, from_type="pd_DataFrame_Table", to_type="polars_eager_table")


@pytest.fixture
def polars_load_diabetes_polars(polars_load_diabetes_pandas):
    X_train, X_test, y_train = polars_load_diabetes_pandas
    X_train_pl = _pd_to_pl(X_train)
    X_test_pl = _pd_to_pl(X_test)
    y_train_pl = _pd_to_pl(y_train)

    # drop the index in the polars frame
    X_train_pl = X_train_pl.drop(["__index__"])
    X_test_pl = X_test_pl.drop(["__index__"])
    y_train_pl = y_train_pl.drop(["__index__"])

    return [X_train_pl, X_test_pl, y_train_pl]


def polars_load_diabetes_polars_with_index(polars_load_diabetes_pandas):
    X_train, X_test, y_train = polars_load_diabetes_pandas
    X_train_pl = _pd_to_pl(X_train)
    X_test_pl = _pd_to_pl(X_test)
    y_train_pl = _pd_to_pl(y_train)

    return [X_train_pl, X_test_pl, y_train_pl]


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes")
    or not _check_soft_dependencies(["polars", "pyarrow"], severity="none"),
    reason="skip test if polars/pyarrow is not installed in environment",
)
def test_polars_eager_conversion_methods(
    polars_load_diabetes_pandas, polars_load_diabetes_polars
):
    """
    Tests to ensure that given a pandas dataframe, the conversion methods
    convert properly to polars dataframe
    """

    X_train, X_test, y_train = polars_load_diabetes_pandas
    X_train_pl, X_test_pl, y_train_pl = polars_load_diabetes_polars

    assert check_is_mtype(X_train_pl, "polars_eager_table")
    assert check_is_mtype(X_test_pl, "polars_eager_table")
    assert check_is_mtype(y_train_pl, "polars_eager_table")

    assert (X_train.values == X_train_pl.to_numpy()).all()
    assert (X_test.values == X_test_pl.to_numpy()).all()
    assert (y_train.values == y_train_pl.to_numpy()).all()


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes")
    or not _check_soft_dependencies(["polars", "pyarrow"], severity="none"),
    reason="skip test if polars/pyarrow is not installed in environment",
)
def test_polars_eager_regressor_in_fit_predict(
    estimator, polars_load_diabetes_pandas, polars_load_diabetes_polars
):
    """
    Tests to ensure that given a polars dataframe, the regression estimator
    can fit and predict and return the correct set of outputs

    Parameters
    ----------

    estimator: a given regression estimator

    """
    # TODO - expand estimator to include a list of regression models to test
    # create a copy of estimator to run further checks
    estimator_copy = estimator
    X_train, X_test, y_train = polars_load_diabetes_pandas
    X_train_pl, X_test_pl, y_train_pl = polars_load_diabetes_polars

    estimator.fit(X_train_pl, y_train_pl)
    y_pred = estimator.predict(X_test_pl)

    assert isinstance(y_pred, pl.DataFrame)
    assert y_pred.columns == y_train_pl.columns
    assert y_pred.shape[0] == X_test_pl.shape[0]

    # code to ensure prediction values match up correctly
    estimator_copy.fit(X_train, y_train)
    y_pred_pd = estimator_copy.predict(X_test)
    assert (y_pred_pd.values == y_pred.to_numpy()).all()


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes")
    or not _check_soft_dependencies(["polars", "pyarrow"], severity="none"),
    reason="skip test if polars/pyarrow is not installed in environment",
)
def test_polars_eager_regressor_in_predict_interval(
    estimator, polars_load_diabetes_polars
):
    X_train_pl, X_test_pl, y_train_pl = polars_load_diabetes_polars
    # TODO - expand estimator to include a list of regression models to test
    estimator.fit(X_train_pl, y_train_pl)
    y_pred_interval = estimator.predict_interval(X_test_pl)

    assert isinstance(y_pred_interval, pd.DataFrame)
    assert y_pred_interval.columns[0] == ("target", 0.9, "lower")
    assert y_pred_interval.columns[1] == ("target", 0.9, "upper")


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes")
    or not _check_soft_dependencies(["polars", "pyarrow"], severity="none"),
    reason="skip test if polars/pyarrow is not installed in environment",
)
def test_polars_eager_regressor_in_predict_quantiles(
    estimator, polars_load_diabetes_polars
):
    X_train_pl, X_test_pl, y_train_pl = polars_load_diabetes_polars

    estimator.fit(X_train_pl, y_train_pl)
    y_pred_quantile = estimator.predict_quantiles(X_test_pl, alpha=TEST_ALPHAS)

    assert isinstance(y_pred_quantile, pd.DataFrame)
    assert y_pred_quantile.columns[0] == ("target", 0.05)
    assert y_pred_quantile.columns[1] == ("target", 0.1)
    assert y_pred_quantile.columns[2] == ("target", 0.25)



"""Metrics for time-to-event or survival prediction."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

__all__ = [
    "ConcordanceHarrell",
    "SPLL",
]

from skpro.metrics.survival._c_harrell import ConcordanceHarrell
from skpro.metrics.survival._spll import SPLL



"""Base classes for adapting probabilistic regressors to the skproframework."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["_DelegateWithFittedParamForwarding"]

from skpro.regression.base.adapters._sklearn import _DelegateWithFittedParamForwarding



"""Implements pipelines for probabilistic supervised regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# based on sktime pipelines

__author__ = ["fkiraly"]
__all__ = ["Pipeline"]

import pandas as pd
from sklearn import clone

from skpro.base import BaseMetaEstimator
from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class _Pipeline(BaseMetaEstimator, BaseProbaRegressor):
    """Abstract class for pipelines."""

    # for default get_params/set_params from BaseMetaEstimator
    # named_object_parameters points to the attribute of self
    # which contains the heterogeneous set of estimators
    # this must be an iterable of (name: str, estimator, ...) tuples for the default
    # if the estimator is fittable, _BaseMetaEstimator also
    # provides an override for get_fitted_params for params from the fitted estimators
    # the fitted estimators should be in fitted_named_object_parameters
    # this must be an iterable of (name: str, estimator, ...) tuples for the default
    _tags = {
        "named_object_parameters": "_steps",
        "fitted_named_object_parameters": "steps_",
    }

    def _get_pipeline_scitypes(self, estimators):
        """Get list of scityes (str) from names/estimator list."""

        def est_scitype(tpl):
            est = tpl[1]
            if isinstance(est, BaseProbaRegressor):
                return "regressor"
            elif hasattr(est, "transform"):
                return "transformer"
            else:
                return "other"

        return [est_scitype(x) for x in estimators]

    def _get_regressor_index(self, estimators):
        """Get the index of the first regressor in the list."""
        return self._get_pipeline_scitypes(estimators).index("regressor")

    def _check_steps(self, estimators, allow_postproc=False):
        """Check Steps.

        Parameters
        ----------
        estimators : list of estimators, or list of (name, estimator) pairs
        allow_postproc : bool, optional, default=False
            whether transformers after the regressor are allowed

        Returns
        -------
        step : list of (name, estimator) pairs, estimators are cloned (not references)
            if estimators was a list of (str, estimator) tuples, then just cloned
            if was a list of estimators, then str are generated via _get_estimator_names

        Raises
        ------
        TypeError if names in `estimators` are not unique
        TypeError if estimators in `estimators` are not all regressor or transformer
        TypeError if there is not exactly one regressor in `estimators`
        TypeError if not allow_postproc and regressor is not last estimator
        """
        if not isinstance(estimators, list):
            msg = (
                f"steps in {self.name} must be list of estimators, "
                f"or (string, estimator) pairs, "
                f"the two can be mixed; but, found steps of type {type(estimators)}"
            )
            raise TypeError(msg)

        estimator_tuples = self._coerce_to_named_object_tuples(estimators)
        estimator_tuples = [(name, clone(est)) for name, est in estimator_tuples]

        def _set_pd(est):
            if hasattr(est, "transform") and hasattr(est, "set_output"):
                est.set_output(transform="pandas")
            return est

        # ensure sklearn transformers produce pandas output
        estimator_tuples = [(name, _set_pd(est)) for name, est in estimator_tuples]

        names, estimators = zip(*estimator_tuples)

        # validate names
        self._check_names(names)

        scitypes = self._get_pipeline_scitypes(estimator_tuples)
        if not set(scitypes).issubset(["regressor", "transformer"]):
            raise TypeError(
                f"estimators passed to {self.name} "
                f"must be either transformer or regressor"
            )
        if scitypes.count("regressor") != 1:
            raise TypeError(
                f"exactly one regressor must be contained in the chain, "
                f"but found {scitypes.count('regressor')}"
            )

        regressor_ind = self._get_regressor_index(estimator_tuples)

        if not allow_postproc and regressor_ind != len(estimators) - 1:
            TypeError(
                f"in {self.name}, last estimator must be a regressor, "
                f"but found a transformer"
            )

        # Shallow copy
        return estimator_tuples

    def _iter_transformers(self, reverse=False, fc_idx=-1):
        # exclude final regressor
        steps = self.steps_[:fc_idx]

        if reverse:
            steps = reversed(steps)

        for idx, (name, transformer) in enumerate(steps):
            yield idx, name, transformer

    def __len__(self):
        """Return the length of the Pipeline."""
        return len(self.steps)

    @property
    def named_steps(self):
        """Map the steps to a dictionary."""
        return dict(self._steps)

    @property
    def _steps(self):
        return self._coerce_to_named_object_tuples(self.steps, clone=False)

    @_steps.setter
    def _steps(self, value):
        self.steps = value

    def _components(self, base_class=None):
        """Return references to all state changing BaseObject type attributes.

        This *excludes* the blue-print-like components passed in the __init__.

        Caution: this method returns *references* and not *copies*.
            Writing to the reference will change the respective attribute of self.

        Parameters
        ----------
        base_class : class, optional, default=None, must be subclass of BaseObject
            if None, behaves the same as `base_class=BaseObject`
            if not None, return dict collects descendants of `base_class`

        Returns
        -------
        dict with key = attribute name, value = reference to attribute
        dict contains all attributes of `self` that inherit from `base_class`, and:
            whose names do not contain the string "__", e.g., hidden attributes
            are not class attributes, and are not hyper-parameters (`__init__` args)
        """
        import inspect

        from skpro.base import BaseEstimator

        if base_class is None:
            base_class = BaseEstimator
        if base_class is not None and not inspect.isclass(base_class):
            raise TypeError(f"base_class must be a class, but found {type(base_class)}")
        # if base_class is not None and not issubclass(base_class, BaseObject):
        #     raise TypeError("base_class must be a subclass of BaseObject")

        fitted_estimator_tuples = self.steps_

        comp_dict = {name: comp for (name, comp) in fitted_estimator_tuples}
        return comp_dict

    # both children use the same step params for testing, so putting it here
    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skbase.utils.dependencies import _check_estimator_deps
        from sklearn.preprocessing import StandardScaler

        from skpro.regression.residual import ResidualDouble
        from skpro.survival.coxph import CoxPH

        params = []

        regressor = ResidualDouble.create_test_instance()

        STEPS1 = [
            ("transformer", StandardScaler()),
            ("regressor", regressor),
        ]
        params += [{"steps": STEPS1}]

        params += [{"steps": [StandardScaler(), regressor]}]

        # testing with survival predictor
        if _check_estimator_deps(CoxPH, severity="none"):
            params += [{"steps": [StandardScaler(), CoxPH()]}]

        return params


class Pipeline(_Pipeline):
    """Pipeline for probabilistic supervised regression.

    Pipeline is only applying the given transformers
    to X. The regressor can also be a TransformedTargetregressor containing
    transformers to transform y.

    For a list `t1`, `t2`, ..., `tN`, `r`
        where `t[i]` are transformers, and `r` is an sktime regressor,
        the pipeline behaves as follows:

    `fit(X, y)` - changes state by running `t1.fit_transform` with `X=X`, `y=y`
        then `t2.fit_transform` on `X=` the output of `t1.fit_transform`, `y=y`, etc
        sequentially, with `t[i]` receiving the output of `t[i-1]` as `X`,
        then running `r.fit` with `X` being the output of `t[N]`, and `y=y`
    `predict(X)` - result is of executing `r.predict`, with `X=X`
        being the result of the following process:
        running `t1.fit_transform` with `X=X`,
        then `t2.fit_transform` on `X=` the output of `t1.fit_transform`, etc
        sequentially, with `t[i]` receiving the output of `t[i-1]` as `X`,
        and returning th output of `tN` to pass to `r.predict` as `X`.
    `predict_interval(X)`, `predict_quantiles(X)` - as `predict(X)`,
        with `predict_interval` or `predict_quantiles` substituted for `predict`
    `predict_var`, `predict_proba` - uses base class default to obtain
        crude estimates from `predict_quantiles`.

    `get_params`, `set_params` uses `sklearn` compatible nesting interface
        if list is unnamed, names are generated as names of classes
        if names are non-unique, `f"_{str(i)}"` is appended to each name string
            where `i` is the total count of occurrence of a non-unique string
            inside the list of names leading up to it (inclusive)

    `Pipeline` can also be created by using the magic multiplication
        on any regressor, i.e., if `my_regressor` inherits from `BaseProbaRegressor`,
            and `my_t1`, `my_t2`, are an `sklearn` transformer,
            then, for instance, `my_t1 * my_t2 * my_regressor`
            will result in the same object as  obtained from the constructor
            `Pipeline([my_t1, my_t2, my_regressor])`
        magic multiplication can also be used with (str, transformer) pairs,
            as long as one element in the chain is a regressor

    Parameters
    ----------
    steps : list of sktime transformers and regressors, or
        list of tuples (str, estimator) of sktime transformers or regressors
            the list must contain exactly one regressor
        these are "blueprint" transformers resp regressors,
            regressor/transformer states do not change when `fit` is called

    Attributes
    ----------
    steps_ : list of tuples (str, estimator) of sktime transformers or regressors
        clones of estimators in `steps` which are fitted in the pipeline
        is always in (str, estimator) format, even if `steps` is just a list
        strings not passed in `steps` are replaced by unique generated strings
        i-th transformer in `steps_` is clone of i-th in `steps`
    regressor_ : estimator, reference to the unique regressor in steps_

    Examples
    --------
    >>> from skpro.regression.compose import Pipeline
    >>> from skpro.regression.residual import ResidualDouble
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.impute import SimpleImputer as Imputer
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.preprocessing import MinMaxScaler
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>>
    >>> reg_mean = LinearRegression()
    >>> reg_proba = ResidualDouble(reg_mean)

    Example 1: string/estimator pairs

    >>> pipe = Pipeline(steps=[
    ...     ("imputer", Imputer()),
    ...     ("scaler", MinMaxScaler()),
    ...     ("regressor", reg_proba),
    ... ])
    >>> pipe.fit(X_train, y_train)
    Pipeline(...)
    >>> y_pred = pipe.predict(X=X_test)
    >>> y_pred_proba = pipe.predict_proba(X=X_test)

    Example 2: without strings

    >>> pipe = Pipeline([
    ...     Imputer(),
    ...     MinMaxScaler(),
    ...     ("regressor", reg_proba),
    ... ])

    Example 3: using the dunder method
    (requires bracketing as sklearn does not support dunders)

    >>> reg_proba = ResidualDouble(reg_mean)
    >>> pipe = Imputer() * (MinMaxScaler() * reg_proba)
    """

    _tags = {
        "capability:multioutput": True,
        "capability:missing": True,
    }

    def __init__(self, steps):
        self.steps = steps
        self.steps_ = self._check_steps(steps, allow_postproc=False)

        super().__init__()

        tags_to_clone = [
            "capability:multioutput",
            "capability:survival",
            "capability:update",
        ]
        self.clone_tags(self.regressor_, tags_to_clone)

    @property
    def regressor_(self):
        """Return reference to the regressor in the pipeline.

        Valid after _fit.
        """
        return self.steps_[-1][1]

    def __rmul__(self, other):
        """Magic * method, return (left) concatenated Pipeline.

        Implemented for `other` being a transformer, otherwise returns `NotImplemented`.

        Parameters
        ----------
        other: `sktime` transformer, must inherit from BaseTransformer
            otherwise, `NotImplemented` is returned

        Returns
        -------
        Pipeline object,
            concatenation of `other` (first) with `self` (last).
            not nested, contains only non-TransformerPipeline `sktime` steps
        """
        _, ests = zip(*self.steps_)
        names = tuple(self._get_names_and_objects(self.steps)[0])

        if hasattr(other, "transform"):
            new_names = (type(other).__name__,) + names
            new_ests = (other,) + ests
        elif isinstance(other, tuple) and len(other) == 2:
            other_name = other[0]
            other_trafo = other[1]
            new_names = (other_name,) + names
            new_ests = (other_trafo,) + ests
        else:
            return NotImplemented

        # if all the names are equal to class names, we eat them away
        if all(type(x[1]).__name__ == x[0] for x in zip(new_names, new_ests)):
            return Pipeline(steps=list(new_ests))
        else:
            return Pipeline(steps=list(zip(new_names, new_ests)))

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # transform X
        for step_idx, name, transformer in self._iter_transformers():
            t = transformer
            X = t.fit_transform(X=X, y=y)
            if not isinstance(X, pd.DataFrame):
                X = pd.DataFrame(X, index=y.index)
            self.steps_[step_idx] = (name, t)

        # fit regressor
        name, regressor = self.steps_[-1]
        r = regressor.clone()
        r.fit(X, y, C=C)
        self.steps_[-1] = (name, r)

        return self

    def _update(self, X, y, C=None):
        """Update regressor with a new batch of training data.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        X = self._transform(X)
        self.regressor_.update(X=X, y=y, C=C)
        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        X = self._transform(X)
        return self.regressor_.predict(X=X)

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        X = self._transform(X)
        return self.regressor_.predict_quantiles(X=X, alpha=alpha)

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        X = self._transform(X)
        return self.regressor_.predict_interval(X=X, coverage=coverage)

    def _predict_var(self, X):
        """Compute/return variance predictions.

        private _predict_var containing the core logic, called from predict_var

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        X = self._transform(X)
        return self.regressor_.predict_var(X=X)

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        X = self._transform(X)
        return self.regressor_.predict_proba(X=X)

    def _transform(self, X, y=None):
        """Transform data."""
        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        for _, _, transformer in self._iter_transformers():
            if self._has_y_arg(transformer.transform):
                Xt = transformer.transform(X=X, y=y)
            else:
                Xt = transformer.transform(X=X)
            if not isinstance(Xt, pd.DataFrame):
                Xt = pd.DataFrame(Xt, index=X.index)
        return Xt

    def _has_y_arg(self, method):
        """Check if method has y argument."""
        from inspect import signature

        sig = signature(method)
        return "y" in sig.parameters



"""Base class for probabilistic survival regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.base import BaseProbaRegressor

# allowed input mtypes
ALLOWED_MTYPES = [
    "pd_DataFrame_Table",
    "pd_Series_Table",
    "numpy1D",
    "numpy2D",
]


class BaseSurvReg(BaseProbaRegressor):
    """Base class for survival regression models.

    Contains no additional logic, only docstring overrides.
    """

    _tags = {"capability:survival": True}

    def fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        self : reference to self
        """
        super().fit(X=X, y=y, C=C)
        return self

    def update(self, X, y, C=None):
        """Update regressor with a new batch of training data.

        Only estimators with the ``capability:update`` tag (value ``True``)
        provide this method, otherwise the method ignores the call and
        discards the data passed.

        State required:
            Requires state to be "fitted".

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        self : reference to self
        """
        super().update(X=X, y=y, C=C)
        return self



# LEGACY MODULE - TODO: remove or refactor

from skpro.workflow.base import Model


class ModelManager:
    """Model manager"""

    def __init__(self):
        self.models = {}

    def register(self, model, tuning=None, group=None, distinguish_tuning=True):
        model = Model(model, tuning, group)

        self.models[model.identifier(distinguish_tuning)] = model

    def group(self, name):
        if isinstance(name, (tuple, list)):
            models = []
            for group in name:
                models = models + self.group(group)
            return models

        models = []
        for k, v in self.models.items():
            if v.group == name:
                models.append(v.clone())

        return models

    def all(self):
        return [v.clone() for v in self.models.values()]

    def info(self):
        print("Total number of registered models: %i\n" % len(self.models))

        for k, v in self.models.items():
            print(k, str(v))

    def __iter__(self):
        for k, v in self.models.items():
            yield v.clone()

    def __getitem__(self, item):
        if item not in self.models:
            return False

        return self.models[item].clone()

    def __contains__(self, item):
        return item in self.models



"""Git related utilities to identify changed modules."""

__author__ = ["fkiraly"]
__all__ = []

import importlib.util
import inspect
import subprocess
from functools import lru_cache


@lru_cache
def get_module_from_class(cls):
    """Get full parent module string from class.

    Parameters
    ----------
    cls : class
        class to get module string from, e.g., NaiveForecaster

    Returns
    -------
    str : module string, e.g., sktime.forecasting.naive
    """
    module = inspect.getmodule(cls)
    return module.__name__ if module else None


@lru_cache
def get_path_from_module(module_str):
    r"""Get local path string from module string.

    Parameters
    ----------
    module_str : str
        module string, e.g., sktime.forecasting.naive

    Returns
    -------
    str : local path string, e.g., sktime\forecasting\naive.py
    """
    try:
        module_spec = importlib.util.find_spec(module_str)
        if module_spec is None:
            raise ImportError(
                f"Error in get_path_from_module, module '{module_str}' not found."
            )
        module_path = module_spec.origin
        if module_path.endswith("__init__.py"):
            return module_path[:-11]
        return module_path
    except Exception as e:
        raise ImportError(f"Error finding module '{module_str}'") from e


@lru_cache
def is_module_changed(module_str):
    """Check if a module has changed compared to the main branch.

    If a child module has changed, the parent module is considered changed as well.

    Parameters
    ----------
    module_str : str
        module string, e.g., sktime.forecasting.naive
    """
    module_file_path = get_path_from_module(module_str)
    cmd = f"git diff remotes/origin/main -- {module_file_path}"
    try:
        output = subprocess.check_output(cmd, shell=True, text=True, encoding="utf-8")
        return bool(output)
    except subprocess.CalledProcessError:
        return True


@lru_cache
def is_class_changed(cls):
    """Check if a class' parent module has changed compared to the main branch.

    Parameters
    ----------
    cls : class
        class to get module string from, e.g., NaiveForecaster

    Returns
    -------
    bool : True if changed, False otherwise
    """
    module_str = get_module_from_class(cls)
    return is_module_changed(module_str)


def get_changed_lines(file_path, only_indented=True):
    """Get changed or added lines from a file.

    Compares the current branch to the origin-main branch.

    Parameters
    ----------
    file_path : str
        path to file to get changed lines from
    only_indented : bool, default=True
        if True, only indented lines are returned, otherwise all lines are returned;
        more precisely, only changed/added lines starting with a space are returned

    Returns
    -------
    list of str : changed or added lines on current branch
    """
    cmd = f"git diff remotes/origin/main -- {file_path}"

    try:
        # Run 'git diff' command to get the changes in the specified file
        result = subprocess.check_output(cmd, shell=True, text=True)

        # if only indented lines are requested, add space to start_chars
        start_chars = "+"
        if only_indented:
            start_chars += " "

        # Extract the changed or new lines and return as a list of strings
        changed_lines = [
            line.strip() for line in result.split("\n") if line.startswith(start_chars)
        ]
        # remove first character ('+') from each line
        changed_lines = [line[1:] for line in changed_lines]

        return changed_lines

    except subprocess.CalledProcessError:
        return []


def get_packages_with_changed_specs():
    """Get packages with changed or added specs.

    Returns
    -------
    list of str : names of packages with changed or added specs
    """
    return list(_get_packages_with_changed_specs())


@lru_cache
def _get_packages_with_changed_specs():
    """Get packages with changed or added specs.

    Private version of get_packages_with_changed_specs,
    to avoid side effects on the list return.

    Returns
    -------
    tuple of str : names of packages with changed or added specs
    """
    from packaging.requirements import InvalidRequirement, Requirement

    changed_lines = get_changed_lines("pyproject.toml")

    packages = []
    for line in changed_lines:
        if line.find("'") > line.find('"') and line.find('"') != -1:
            sep = '"'
        elif line.find("'") == -1:
            sep = '"'
        else:
            sep = "'"

        splits = line.split(sep)
        if len(splits) < 2:
            continue

        req = line.split(sep)[1]

        # deal with ; python_version >= "3.7" in requirements
        if ";" in req:
            req = req.split(";")[0]

        try:  # deal with lines that are not package requirement strings
            pkg = Requirement(req).name
        except InvalidRequirement:
            continue
        else:
            packages.append(pkg)

    # make unique
    packages = tuple(set(packages))

    return packages



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Common utilities for polars based data containers."""
from skpro.datatypes._common import _req
from skpro.datatypes._common import _ret as ret


def check_polars_frame(obj, return_metadata=False, var_name="obj", lazy=False):
    """Check polars frame, generic format."""
    import polars as pl

    metadata = {}

    if lazy:
        exp_type = pl.LazyFrame
        exp_type_str = "LazyFrame"
    else:
        exp_type = pl.DataFrame
        exp_type_str = "DataFrame"

    if not isinstance(obj, exp_type):
        msg = f"{var_name} must be a polars {exp_type_str}, found {type(obj)}"
        return ret(False, msg, None, return_metadata)

    # we now know obj is a polars DataFrame or LazyFrame
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = obj.width < 1
    if _req("is_univariate", return_metadata):
        obj_width = obj.width
        for col in obj.columns:
            if "__index__" in col:
                obj_width -= 1
        metadata["is_univariate"] = obj_width == 1
    if _req("n_instances", return_metadata):
        if hasattr(obj, "height"):
            metadata["n_instances"] = obj.height
        else:
            metadata["n_instances"] = "NA"
    if _req("n_features", return_metadata):
        obj_width = obj.width
        for col in obj.columns:
            if "__index__" in col:
                obj_width -= 1
        metadata["n_features"] = obj_width
    if _req("feature_names", return_metadata):
        if lazy:
            obj_columns = obj.collect_schema().names()
            feature_names = [
                col for col in obj_columns if not col.startswith("__index__")
            ]
            metadata["feature_names"] = feature_names
        else:
            obj_columns = obj.columns
            feature_names = [
                col for col in obj_columns if not col.startswith("__index__")
            ]
            metadata["feature_names"] = feature_names

    # check if there are any nans
    #   compute only if needed
    if _req("has_nans", return_metadata):
        if isinstance(obj, pl.LazyFrame):
            metadata["has_nans"] = "NA"
        else:
            hasnan = obj.null_count().sum_horizontal().to_numpy()[0] > 0
            metadata["has_nans"] = hasnan

    return ret(True, None, metadata, return_metadata)


def convert_polars_to_pandas_with_index(obj):
    """Convert function from polars to pandas,converts  __index__ to pandas index.

    Parameters
    ----------
    obj : polars DataFrame, polars.LazyFrame

    Returns
    -------
    pd_df : pandas DataFrame
        Returned is a pandas DataFrame with index retained if column __index__
        existed in the polars dataframe previously, if not then index of
        pd_df will be a RangeIndex from 0 to pd_df.shape[0]-1.

    """
    from polars.lazyframe.frame import LazyFrame

    if isinstance(obj, LazyFrame):
        obj = obj.collect()

    pd_df = obj.to_pandas()
    for col in obj.columns:
        if col.startswith("__index__"):
            pd_df = pd_df.set_index(col, drop=True)
            pd_df.index.name = col.split("__index__")[1]

    return pd_df


def convert_pandas_to_polars_with_index(
    obj, schema_overrides=None, rechunk=True, nan_to_null=True, lazy=False
):
    """Convert function from pandas to polars, and preserves index.

    Parameters
    ----------
    obj : pandas DataFrame

    schema_overrides : dict, optional (default=None)
        Support override of inferred types for one or more columns.

    rechunk : bool, optional (default=True)
        Make sure that all data is in contiguous memory.

    nan_to_null : bool, optional (default=True)
        If data contains NaN values PyArrow will convert the NaN to None

    lazy : bool, optional (default=False)
        If True, return a LazyFrame instead of a DataFrame

    Returns
    -------
    pl_df : polars DataFrame or polars LazyFrame
        index from pandas DataFrame will be returned as a polars column
        named __index__.
    """
    import pandas as pd
    from polars import from_pandas

    # if the index of the dataframe is the trivial index (i.e RangeIndex(0,numrows))
    # we do not return an __index__ column
    if not (
        isinstance(obj.index, pd.RangeIndex)
        and obj.index.start == 0
        and obj.index.stop == len(obj)
    ):
        obj_index_name = obj.index.name
        obj = obj.reset_index()
        if obj_index_name is not None:
            obj = obj.rename(columns={obj_index_name: f"__index__{obj_index_name}"})
        else:
            obj = obj.rename(columns={"index": "__index__"})

    pl_df = from_pandas(
        data=obj,
        schema_overrides=schema_overrides,
        rechunk=rechunk,
        nan_to_null=nan_to_null,
    )

    if lazy:
        pl_df = pl_df.lazy()

    return pl_df
